/**CHeaderFile*************************************************************
  PackageName [Biddy]
  Synopsis    [Biddy provides data structures and algorithms
               for the representation and manipulation of Boolean
               functions with ROBDDs. A hash table is used for quick
               search of nodes. Complement edges decreases the number
               of nodes. Function ITE is given for manipulation of
               boolean functions. An automatic garbage collection with
               a formulae counter is implemented.]

  FileName    [bdd.h]
  Revision    [$Revision: 16 $]
  Date        [$Date: 2007-07-19 14:10:30 +0200 (Thu, 19 Jul 2007) $]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Ales Casar (casar@uni-mb.si)]
  Description [The file bdd.h contains declaration of all external
               data structures.]
  SeeAlso     [bddInt.h]

  Copyright   [This file is part of Biddy.
               Copyright (C) 2006,2007
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               Biddy is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               Biddy is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

#ifndef _BIDDY
#define _BIDDY

#define BIDDY_VERSION BIDDYVERSION

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#include <unistd.h>
#include <math.h>

/*-----------------------------------------------------------------------*/
/* Constant definitions                                                  */
/*-----------------------------------------------------------------------*/

#define FALSE 0
#define TRUE 1

/*-----------------------------------------------------------------------*/
/* Macro definitions                                                     */
/*-----------------------------------------------------------------------*/

/* Biddy_Eqv return TRUE iff given BDDs are equal */

#define Biddy_isEqv(f,g) (((f).p == (g).p) && ((f).mark == (g).mark))

/* Biddy_Terminal return TRUE iff given BDD is a constant */

#define Biddy_isTerminal(f) ((f).p == biddy_termTrue.p)

/* Biddy_Null return TRUE iff given BDD is a NULL edge */

#define Biddy_isNull(f) ((f).p == biddy_termNull.p)

/*-----------------------------------------------------------------------*/
/* Type declarations                                                     */
/*-----------------------------------------------------------------------*/

typedef char Biddy_Boolean;
typedef char *Biddy_String;
typedef int Biddy_Variable;
typedef void (*Biddy_UserFunction)();

/*-----------------------------------------------------------------------*/
/* Structure declarations                                                */
/*-----------------------------------------------------------------------*/

/* EDGE */
typedef struct Biddy_Edge {
  void *p;
  Biddy_Boolean mark;
} Biddy_Edge;

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

extern Biddy_Edge biddy_termTrue;            /* terminal node */

extern Biddy_Edge biddy_termFalse;           /* terminal node */

extern Biddy_Edge biddy_termNull;                /* null edge */

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Function prototypes                                                   */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* FUNCTIONS EXPORTED FROM biddyMain.c                                   */
/*-----------------------------------------------------------------------*/

extern void Biddy_Init();

extern void Biddy_Exit();

extern void Biddy_About();

extern Biddy_Boolean Biddy_isOK(Biddy_Edge f);

extern Biddy_Edge Biddy_GetThen(Biddy_Edge f);

extern Biddy_Edge Biddy_GetElse(Biddy_Edge f);

extern Biddy_Variable Biddy_GetVariable(Biddy_Edge f);

extern Biddy_String Biddy_GetVariableName(Biddy_Edge f);

extern char Biddy_GetVariableChar(Biddy_Edge f);

extern Biddy_String Biddy_GetIthVariableName(int i);

extern int Biddy_GetVariableOrder(Biddy_Edge f);

extern void Biddy_SelectNode(Biddy_Edge f);

extern void Biddy_DeselectNode(Biddy_Edge f);

extern Biddy_Boolean Biddy_isSelected(Biddy_Edge f);

extern Biddy_Edge Biddy_FindTerminal(Biddy_String x);

extern Biddy_Edge Biddy_FoaTerminal(Biddy_String x);

extern Biddy_Edge Biddy_FoaNode(Biddy_Variable v, Biddy_Edge pf, Biddy_Edge pt);

extern Biddy_Edge Biddy_NOT(Biddy_Edge f);

extern Biddy_Edge Biddy_TransferMark(Biddy_Edge f, Biddy_Boolean mark);

extern Biddy_Edge Biddy_ITE(Biddy_Edge f, Biddy_Edge g, Biddy_Edge h);

extern Biddy_Edge Biddy_Restrict(Biddy_Edge f, Biddy_Variable v, Biddy_Boolean value);

extern Biddy_Edge Biddy_Compose(Biddy_Edge f, Biddy_Variable v, Biddy_Edge g);

extern Biddy_Edge Biddy_E(Biddy_Edge f, Biddy_Variable v);

extern Biddy_Edge Biddy_A(Biddy_Edge f, Biddy_Variable v);

extern void Biddy_IncCounter();

extern void Biddy_Fresh(Biddy_Edge f);

extern void Biddy_Fortify(Biddy_Edge f);

extern void Biddy_Garbage();

extern void Biddy_AddCache(Biddy_UserFunction gc);

extern void Biddy_NodeSelect(Biddy_Edge f);

extern void Biddy_NodeRepair(Biddy_Edge f);

/*-----------------------------------------------------------------------*/
/* FUNCTIONS EXPORTED FROM biddyStat.c                                   */
/*-----------------------------------------------------------------------*/

extern int Biddy_NodeTableSize();

extern int Biddy_NodeTableMax();

extern int Biddy_NodeTableNum();

extern int Biddy_NodeTableNumF();

extern int Biddy_NodeTableFOA();

extern int Biddy_NodeTableCompare();

extern int Biddy_NodeTableAdd();

extern int Biddy_NodeTableGarbage();

extern int Biddy_NodeTableGenerated();

extern int Biddy_NodeTableBlockNumber();

extern int Biddy_ListUsed();

extern int Biddy_ListMaxLength();

extern float Biddy_ListAvgLength();

extern int Biddy_IteCacheFind();

extern int Biddy_IteCacheOverwrite();

extern int Biddy_NodeNumber(Biddy_Edge f);

extern int Biddy_NodeMaxLevel(Biddy_Edge f);

extern float Biddy_NodeAvgLevel(Biddy_Edge f);

/**AutomaticEnd***********************************************************/

#endif  /* _BIDDY */
