/**CFile*******************************************************************
  PackageName [bdd]
  Synopsis    [Package 'bdd' enable symbolic computations by representing
               Boolean functions with ROBDDs.]

  FileName    [bddMain.c]
  Revision    [$Revision: 13 $]
  Date        [$Date: 2007-07-19 14:02:50 +0200 (Thu, 19 Jul 2007) $]
  Authors     [Robert Meolic (meolic@uni-mb.si)]
  Description [File bddMain.c contains main functions for
               representation and manipulation of boolean functions
               with ROBDDs.]
  SeeAlso     [bdd.h, bddInt.h]

  Copyright   [This file is part of EST (Efficient Symbolic Tools).
               Copyright (C) 2007
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               EST is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               EST is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

#include <biddy.h>
#include "bddInt.h"

/*-----------------------------------------------------------------------*/
/* Constant declarations                                                 */
/*-----------------------------------------------------------------------*/

#define SMALL_RELOP_CACHE  1009
#define MEDIUM_RELOP_CACHE 10007
#define BIG_RELOP_CACHE    100043
#define HUGE_RELOP_CACHE   500057

#define SMALL_RELOPCOMPLEX_CACHE  1009
#define MEDIUM_RELOPCOMPLEX_CACHE 10007
#define BIG_RELOPCOMPLEX_CACHE    100043
#define HUGE_RELOPCOMPLEX_CACHE   500057


/*-----------------------------------------------------------------------*/
/* Macro definitions                                                     */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/* EXPORTED VARIABLES */

int bdd_status = 0;                             /* initialization status */

Bdd_Edge bdd_termTrue;                                  /* terminal node */
Bdd_Edge bdd_termFalse;                                 /* terminal node */
Bdd_Edge bdd_termNull;                                      /* null edge */

/* INTERNAL VARIABLES */

unsigned int bddRelOpCacheSize = HUGE_RELOP_CACHE;
unsigned int bddRelOpComplexCacheSize = HUGE_RELOPCOMPLEX_CACHE;

BddFormula *bddFormulaeTree;                            /* formulae Tree */
BddRelOpCacheTable bddRelOpCache;                          /* user cache */
BddRelOpComplexCacheTable bddRelOpComplexCache;            /* user cache */

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static void RelOpGarbage();

static void RelOpComplexGarbage();

static void deleteFormulaeTree(BddFormula *f);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

#ifndef TCLOUT
   void Bdd_Init(void) {}
#endif

/**Function****************************************************************
  Synopsis    [Bdd_isOK]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_isOK]
  CUDD        []
  ************************************************************************/

Est_Boolean Bdd_isOK(Bdd_Edge f)
{
  Biddy_Edge ff;

  ff.p = f.p;
  ff.mark = f.mark;
  return (Est_Boolean) Biddy_isOK(ff);
}

/**Function****************************************************************
  Synopsis    [Bdd_GetThen]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_GetThen]
  CUDD        [Cudd_T]
  ************************************************************************/

Bdd_Edge Bdd_GetThen(Bdd_Edge f)
{

  /* BIDDY */
  /**/
  Biddy_Edge ff,rr;
  Bdd_Edge r;

  ff.p = f.p;
  ff.mark = f.mark;
  rr = Biddy_GetThen(ff);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/

  /* CUDD */
  /*
  Bdd_Edge r;
  r.p = (void *) Cudd_T((DdNode) f.p);
  r.mark = Cudd_IsComplement((DdNode) r.p);
  return r;
  */

}

/**Function****************************************************************
  Synopsis    [Bdd_GetElse]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_GetElse]
  CUDD        [Cudd_E]
  ************************************************************************/

Bdd_Edge Bdd_GetElse(Bdd_Edge f)
{

  /* BIDDY */
  /**/
  Biddy_Edge ff,rr;
  Bdd_Edge r;

  ff.p = f.p;
  ff.mark = f.mark;
  rr = Biddy_GetElse(ff);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/

  /* CUDD */
  /*
  Bdd_Edge r;
  r.p = (void *) Cudd_E((DdNode) f.p);
  r.mark = Cudd_IsComplement((DdNode) r.p);
  return r;
  */
}

/**Function****************************************************************
  Synopsis    [Bdd_GetVariable]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_GetVariable]
  CUDD        [-]
  ************************************************************************/

Bdd_Variable Bdd_GetVariable(Bdd_Edge f)
{

  /* BIDDY */
  /**/
  Biddy_Edge ff;

  ff.p = f.p;
  ff.mark = f.mark;
  return (Bdd_Variable) Biddy_GetVariable(ff);
  /**/

  /* CUDD */
  /*
  return (Bdd_Variable) (Cudd_Regular((DdNode) r.p))->index;
  */
}

/**Function****************************************************************
  Synopsis    [Bdd_GetVariableName]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

Est_String Bdd_GetVariableName(Bdd_Edge f)
{

  /* BIDDY */
  /**/
  Biddy_Edge ff;

  ff.p = f.p;
  ff.mark = f.mark;
  return (Est_String) Biddy_GetVariableName(ff);
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_GetVariableChar]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

char Bdd_GetVariableChar(Bdd_Edge f)
{

  /* BIDDY */
  /**/
  Biddy_Edge ff;

  ff.p = f.p;
  ff.mark = f.mark;
  return Biddy_GetVariableChar(ff);
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_GetVariableName]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

Est_String Bdd_GetIthVariableName(int i)
{

  /* BIDDY */
  /**/
  return (Est_String) Biddy_GetIthVariableName(i);
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_GetVariableName]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

int Bdd_GetVariableOrder(Bdd_Edge f)
{

  /* BIDDY */
  /**/
  Biddy_Edge ff;

  ff.p = f.p;
  ff.mark = f.mark;
  return Biddy_GetVariableOrder(ff);
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_FindTerminal]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_FindTerminal]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_FindTerminal(Est_String x)
{

  /* BIDDY */
  /**/
  Biddy_Edge rr;
  Bdd_Edge r;

  rr = Biddy_FindTerminal((Biddy_String) x);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_FoaTerminal]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_FoaTerminal]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_FoaTerminal(Est_String x)
{

  /* BIDDY */
  /**/
  Biddy_Edge rr;
  Bdd_Edge r;

  rr = Biddy_FoaTerminal((Biddy_String) x);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_FoaNode]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_FoaNode]
  CUDD        [cuddUniqueInter]
  ************************************************************************/

Bdd_Edge Bdd_FoaNode(Bdd_Variable v, Bdd_Edge pf, Bdd_Edge pt)
{

  /* BIDDY */
  /**/
  Biddy_Edge rr,pff,ptt;
  Bdd_Edge r;

  pff.p = pf.p;
  pff.mark = pf.mark;
  ptt.p = pt.p;
  ptt.mark = pt.mark;
  rr = Biddy_FoaNode((Biddy_Variable) v, pff, ptt);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_TransferMark]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_TransferMark]
  CUDD        [Cudd_NotCond]
  ************************************************************************/

Bdd_Edge Bdd_TransferMark(Bdd_Edge f, Est_Boolean mark)
{

  /* BIDDY */
  /**/
  Biddy_Edge rr,ff;
  Bdd_Edge r;

  ff.p = f.p;
  ff.mark = f.mark;
  rr = Biddy_TransferMark(ff, (Biddy_Boolean) mark);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_NOT]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NOT]
  CUDD        [Cudd_Not]
  ************************************************************************/

Bdd_Edge Bdd_NOT(Bdd_Edge f)
{

  /* BIDDY */
  /**/
  Biddy_Edge rr,ff;
  Bdd_Edge r;

  ff.p = f.p;
  ff.mark = f.mark;
  rr = Biddy_NOT(ff);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_ITE]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_ITE]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_ITE(Bdd_Edge f, Bdd_Edge g, Bdd_Edge h)
{

  /* BIDDY */
  /**/
  Biddy_Edge rr,ff,gg,hh;
  Bdd_Edge r;

  ff.p = f.p;
  ff.mark = f.mark;
  gg.p = g.p;
  gg.mark = g.mark;
  hh.p = h.p;
  hh.mark = h.mark;
  rr = Biddy_ITE(ff,gg,hh);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_Restrict]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Restrict]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_Restrict(Bdd_Edge f, Bdd_Variable v, Est_Boolean value)
{

  /* BIDDY */
  /**/
  Biddy_Edge rr,ff;
  Bdd_Edge r;

  ff.p = f.p;
  ff.mark = f.mark;
  rr = Biddy_Restrict(ff, (Biddy_Variable) v, (Biddy_Boolean) value);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_Compose]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Compose]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_Compose(Bdd_Edge f, Bdd_Variable v, Bdd_Edge g)
{

  /* BIDDY */
  /**/
  Biddy_Edge rr,ff,gg;
  Bdd_Edge r;

  ff.p = f.p;
  ff.mark = f.mark;
  gg.p = g.p;
  gg.mark = g.mark;
  rr = Biddy_Compose(ff, (Biddy_Variable) v, gg);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_E]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_E]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_E(Bdd_Edge f, Bdd_Variable v)
{

  /* BIDDY */
  /**/
  Biddy_Edge rr,ff;
  Bdd_Edge r;

  ff.p = f.p;
  ff.mark = f.mark;
  rr = Biddy_E(ff, (Biddy_Variable) v);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_A]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_A]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_A(Bdd_Edge f, Bdd_Variable v)
{

  /* BIDDY */
  /**/
  Biddy_Edge rr,ff;
  Bdd_Edge r;

  ff.p = f.p;
  ff.mark = f.mark;
  rr = Biddy_A(ff, (Biddy_Variable) v);

  r.p = rr.p;
  r.mark = rr.mark;
  return r;
  /**/
}

/**Function****************************************************************
  Synopsis    [Bdd_IncCounter]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_IncCounter]
  CUDD        []
  ************************************************************************/

void Bdd_IncCounter()
{
  Biddy_IncCounter();
}

/**Function****************************************************************
  Synopsis    [Bdd_Fresh]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Fresh]
  CUDD        []
  ************************************************************************/

void Bdd_Fresh(Bdd_Edge f)
{
  Biddy_Edge ff;

  ff.p = f.p;
  ff.mark = f.mark;

  Biddy_Fresh(ff);
}

/**Function****************************************************************
  Synopsis    [Bdd_Fortify]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Fortify]
  CUDD        []
  ************************************************************************/

void Bdd_Fortify(Bdd_Edge f)
{
  Biddy_Edge ff;

  ff.p = f.p;
  ff.mark = f.mark;
  Biddy_Fortify(ff);
}

/**Function****************************************************************
  Synopsis    [Bdd_Garbage]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Garbage]
  CUDD        []
  ************************************************************************/

void Bdd_Garbage()
{
  Biddy_Garbage();
}

/**Function****************************************************************
  Synopsis    [Bdd_AddCache]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_AddCache]
  CUDD        []
  ************************************************************************/

void Bdd_AddCache(Bdd_UserFunction gc)
{
  Biddy_AddCache((Biddy_UserFunction) gc);
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeTableSize]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableSize]
  CUDD        []
  ************************************************************************/

int Bdd_NodeTableSize()
{
  return Biddy_NodeTableSize();
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeTableMax]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableMax]
  CUDD        []
  ************************************************************************/

int Bdd_NodeTableMax()
{
  return Biddy_NodeTableMax();
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeTableNum]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableNum]
  CUDD        []
  ************************************************************************/

int Bdd_NodeTableNum()
{
  return Biddy_NodeTableNum();
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeTableNumF]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableNumF]
  CUDD        []
  ************************************************************************/

int Bdd_NodeTableNumF()
{
  return Biddy_NodeTableNumF();
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeTableFOA]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableFOA]
  CUDD        []
  ************************************************************************/

int Bdd_NodeTableFOA()
{
  return Biddy_NodeTableFOA();
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeTableCompare]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableCompare]
  CUDD        []
  ************************************************************************/

int Bdd_NodeTableCompare()
{
  return Biddy_NodeTableCompare();
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeTableAdd]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableAdd]
  CUDD        []
  ************************************************************************/

int Bdd_NodeTableAdd()
{
  return Biddy_NodeTableAdd();
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeTableGarbage]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableGarbage]
  CUDD        []
  ************************************************************************/

int Bdd_NodeTableGarbage()
{
  return Biddy_NodeTableGarbage();
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeTableGenerated]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableGenerated]
  CUDD        []
  ************************************************************************/

int Bdd_NodeTableGenerated()
{
  return Biddy_NodeTableGenerated();
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeTableBlockNumber]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableBlockNumber]
  CUDD        []
  ************************************************************************/

int Bdd_NodeTableBlockNumber()
{
  return Biddy_NodeTableBlockNumber();
}

/**Function****************************************************************
  Synopsis    [Bdd_ListUsed]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_ListUsed]
  CUDD        []
  ************************************************************************/

int Bdd_ListUsed()
{
  return Biddy_ListUsed();
}

/**Function****************************************************************
  Synopsis    [Bdd_ListMaxLength]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_ListMaxLength]
  CUDD        []
  ************************************************************************/

int Bdd_ListMaxLength()
{
  return Biddy_ListMaxLength();
}

/**Function****************************************************************
  Synopsis    [Bdd_ListAvgLength]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_ListAvgLength]
  CUDD        []
  ************************************************************************/

float Bdd_ListAvgLength()
{
  return Biddy_ListAvgLength();
}

/**Function****************************************************************
  Synopsis    [Bdd_IteCacheFind]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_IteCacheFind]
  CUDD        []
  ************************************************************************/

int Bdd_IteCacheFind()
{
  return Biddy_IteCacheFind();
}

/**Function****************************************************************
  Synopsis    [Bdd_IteCacheOverwrite]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_IteCacheOverwrite]
  CUDD        []
  ************************************************************************/

int Bdd_IteCacheOverwrite()
{
  return Biddy_IteCacheOverwrite();
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeNumber]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeNumber]
  CUDD        []
  ************************************************************************/

int Bdd_NodeNumber(Bdd_Edge f)
{
  Biddy_Edge ff;

  ff.p = f.p;
  ff.mark = f.mark;
  return Biddy_NodeNumber(ff);
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeMaxLevel]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeMaxLevel]
  CUDD        []
  ************************************************************************/

int Bdd_NodeMaxLevel(Bdd_Edge f)
{
  Biddy_Edge ff;

  ff.p = f.p;
  ff.mark = f.mark;
  return Biddy_NodeMaxLevel(ff);
}

/**Function****************************************************************
  Synopsis    [Bdd_NodeAvgLevel]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Bdd_NodeAvgLevel]
  CUDD        []
  ************************************************************************/

float Bdd_NodeAvgLevel(Bdd_Edge f)
{
  Biddy_Edge ff;

  ff.p = f.p;
  ff.mark = f.mark;
  return Biddy_NodeAvgLevel(ff);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_InitPkg()
{
    int i;

    if (bdd_status == 1) {
      printf("\nBDD package is already initialized. ");
      return;
    } else {
      bdd_status = 1;
    }

    Biddy_Init();

    /* INITIALIZATION OF TERMINAL NODES */
    bdd_termTrue.p = biddy_termTrue.p;
    bdd_termTrue.mark = biddy_termTrue.mark;
    bdd_termFalse.p = biddy_termFalse.p;
    bdd_termFalse.mark = biddy_termFalse.mark;
    bdd_termNull.p = biddy_termNull.p;
    bdd_termNull.mark = biddy_termNull.mark;

    /* INITIALIZATION OF FORMULAE TREE */
    bddFormulaeTree = NULL;
    Bdd_AddFormula("1",bdd_termTrue);
    Bdd_AddFormula("0",bdd_termFalse);

    /* INITIALIZATION OF RelOp CACHE - USED FOR RELATIONAL OPERATIONS */
    bddRelOpCache.size = bddRelOpCacheSize;
    bddRelOpCache.search = 0;
    bddRelOpCache.find = 0;
    bddRelOpCache.write = 0;
    bddRelOpCache.overwrite = 0;
    if (!(bddRelOpCache.table = (BddRelOpCache *)
          malloc(bddRelOpCache.size * sizeof(BddRelOpCache)))) {
      fprintf(stdout,"Bdd_InitPkg (RelOp): Out of memoy!\n");
      exit(1);
    }
    for (i=0;i<bddRelOpCache.size;i++) {
      bddRelOpCache.table[i].key = 0;
    }
    Bdd_AddCache(RelOpGarbage);

    /* INITIALIZATION OF RelOpComplex CACHE - ADVANCED RELATIONAL OPERAT. */
    bddRelOpComplexCache.size = bddRelOpComplexCacheSize;
    bddRelOpCache.search = 0;;
    bddRelOpComplexCache.search = 0;
    bddRelOpComplexCache.find = 0;
    bddRelOpComplexCache.write = 0;
    bddRelOpComplexCache.overwrite = 0;
    if (!(bddRelOpComplexCache.table = (BddRelOpComplexCache *)
          malloc(bddRelOpComplexCache.size * sizeof(BddRelOpComplexCache)))) {
      fprintf(stdout,"Bdd_InitPkg (RelOpComplex): Out of memoy!\n");
      exit(1);
    }
    for (i=0;i<bddRelOpComplexCache.size;i++) {
      bddRelOpComplexCache.table[i].key = 0;
    }
    Bdd_AddCache(RelOpComplexGarbage);

}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_ExitPkg()
{
  int i;

  if (bdd_status == 0) {
    printf("\nBDD package is not initialized. ");
    return;
  } else {
    bdd_status = 0;
  }

  Biddy_Exit();

  /* DELETE FORMULAE TREE */
  /*
  printf("Delete formulae tree...\n");
  */
  deleteFormulaeTree(bddFormulaeTree);

  /* DELETE RelOp CACHE TABLE */
  /*
  printf("Delete RelOp cache...\n");
  */
  free(bddRelOpCache.table);

  /* DELETE RelOpComplex CACHE TABLE */
  /*
  printf("Delete RelOpComplex cache...\n");
  */
  for (i=0; i<bddRelOpComplexCache.size; i++) {
    if (bddRelOpComplexCache.table[i].key) {
      free(bddRelOpComplexCache.table[i].name);
    }
  }
  free(bddRelOpComplexCache.table);

}

/**Function****************************************************************
  Synopsis    []
  Description [Bdd_AboutPkg reports type and version of BDD package.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_AboutPkg()
{
  printf("EST ");
  printf(EDITION);
  printf(", version ");
  printf(VERSION);
  printf(", ");
  printf("Bdd package");
  printf("\n");

  Biddy_About();
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function RelOpGarbage.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
RelOpGarbage()
{
  int i;
  BddRelOpCache *p;

  for (i=0; i<bddRelOpCache.size; i++) {
    if (bddRelOpCache.table[i].key) {
      p = &bddRelOpCache.table[i];
      if (!Bdd_isOK(p->f) || !Bdd_isOK(p->g) || !Bdd_isOK(p->r))
      {
        p->key = 0;
      }
    }
  }
}

/**Function****************************************************************
  Synopsis    [Function RelOpComplexGarbage.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
RelOpComplexGarbage()
{
  int i;
  BddRelOpComplexCache *p;

  for (i=0; i<bddRelOpComplexCache.size; i++) {
    if (bddRelOpComplexCache.table[i].key) {
      p = &bddRelOpComplexCache.table[i];
      if (!Bdd_isOK(p->f) || !Bdd_isOK(p->g) || !Bdd_isOK(p->r))
      {
        free(p->name);
        p->key = 0;
      }
    }
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
deleteFormulaeTree(BddFormula *f)
{
  if (f != NULL) {
    deleteFormulaeTree(f->l);
    deleteFormulaeTree(f->r);
    free(f->name);
    free(f);
  }
}
