/**CFile*******************************************************************
  PackageName [pa]
  Synopsis    [Package 'pa' is a framework for implementation of
               process algebrae similar to Milner's CCS and Hoare's CSP.]

  FileName    [paKod.c]
  Revision    [$Revision: 13 $]
  Date        [$Date: 2007-07-19 14:02:50 +0200 (Thu, 19 Jul 2007) $]
  Authors     [Robert Meolic (meolic@uni-mb.si)]
  Description [File paKod.c implements encoding and decoding of LTSs.]
  SeeAlso     [pa.h, paInt.h]

  Copyright   [This file is part of EST (Efficient Symbolic Tools).
               Copyright (C) 1999, 2003, 2006, 2007
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               EST is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               EST is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

#include <math.h>
#include "paInt.h"

/*-----------------------------------------------------------------------*/
/* Structure declarations                                                */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static Bdd_Edge EncodeMinterm(Bdd_Edge *tabVar, int n, int i, int min);

static int DecodeSortBddA(Pa_Sort *sort, Bdd_Edge F,
                          Est_Boolean mark, int minA, int maxA,
                          Est_Boolean write);

static int DecodeProcessBddR(Pa_Process *proc, Bdd_Edge F,
                             Est_Boolean mark, int minR, int maxR,
                             Est_Boolean write);

static int DecodeProcessBddRAS(Pa_Process *proc, Bdd_Edge F,
                               Est_Boolean mark,
                               int minR, int maxR, int minA, int maxA,
                               int minS, int maxS, Est_Boolean write);

static int DecodeCompBddR(Pa_Composition *comp, Bdd_Edge F,
                          Est_Boolean mark, int *decodingTableR,
                          Est_Boolean write);

static int DecodeCompBddRAS(Pa_Composition *comp, Bdd_Edge F,
                            Est_Boolean mark, int minA, int maxA,
                            int *decodingTableR, int *decodingTableS,
                            Est_Boolean write);

static int DecodeProcProcPairBddRP(
                           Pa_Process *p1, Pa_Process *p2, Bdd_Edge F,
                           Est_Boolean mark, int minR, int maxR,
                           int minP, int maxP, Est_Boolean write);

static int DecodeCompProcPairBddRP(
                           Pa_Composition *comp, Pa_Process *p, Bdd_Edge F,
                           Est_Boolean mark, int *decodingTableR,
                           int minP, int maxP, Est_Boolean write);

static int DecodeCompCompPairBddRP(
                           Pa_Composition *c1, Pa_Composition *c2, Bdd_Edge F,
                           Est_Boolean mark, int *decodingTableR,
                           int *decodingTableS, Est_Boolean write);

static int DecodeSortAction(Pa_Sort *s, int minA, int maxA,
                            Est_Boolean write);

static int DecodeProcessState(Pa_Process *p, int minR, int maxR,
                              Est_Boolean write);

static int DecodeProcessTR(Pa_Process *p, int minR, int maxR,
                           int minA, int maxA, int minS, int maxS,
                           Est_Boolean write);

static int DecodeCompState(Pa_Composition *c, int *decodingTableR,
                           Est_Boolean write);

static int DecodeCompTR(Pa_Composition *c, int minA, int maxA,
                        int *decodingTableR, int *decodingTableS,
                        Est_Boolean write);

static int DecodeProcProcPair(Pa_Process *p1, Pa_Process *p2,
                              int minR, int maxR, int minP, int maxP,
                              Est_Boolean write);

static int DecodeCompProcPair(Pa_Composition *c, Pa_Process *p,
                              int *decodingTableR, int minP, int maxP,
                              Est_Boolean write);

static int DecodeCompCompPair(Pa_Composition *c1, Pa_Composition *c2,
                              int *decodingTableR, int *decodingTableS,
                              Est_Boolean write);

static int DecodeAction(Pa_Sort *s, int min, int max, int f,
                        Est_Boolean write);

static int DecodeState(Pa_Process *p, int min, int max, int f,
                       Est_Boolean write);

static void setproccomp(Pa_Composition *comp);

static int getWeight(Bdd_Edge f);

static int getProcComp(Bdd_Edge f);

/**AutomaticEnd***********************************************************/


/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description [V tabeli sortov poiscemo sort z danim imenom in ga kodiramo.
               Action 'TAU' has to be the first action in the table!]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_EncodeSort(Est_String name)
{
  Pa_Sort *s;
  Bdd_Edge var;
  Bdd_Edge *tabVar;
  Est_String x;
  int i,n,w;

  i = Pa_FindSort(name);
  if (i == -1) return;
  s = &pa_sortTable[i];

  /* dolocimo stevilo potrebnih spremenljivk za kodiranje akcij */
  n = (int)(1.9999999+log(s->numActions+1)/log(2.0));
  s->numVar = n;

  /* tvorimo spremenljivke in dolocimo utezi */
  tabVar = (Bdd_Edge *) malloc(n * sizeof(Bdd_Edge));
  x = (Est_String) malloc(127);
  w = 1;
  for (i=0; i<n; i++) {
    sprintf(x,"a%d",i);            /* format "a%d\0" */
    var = Bdd_FoaTerminal(x);
    Pa_InitVariable(Bdd_GetVariable(var));
    pa_variableTable[Bdd_GetVariable(var)].weight = w;
    w = w * 2;
    tabVar[i] = var;
  }

  free(x);

  /* spremenljivka, ki doloca tip akcije */

  s->a0 = tabVar[0];

  /* MAKE BDD FOR ACTION 'TAU' */

  s->table[0].p = EncodeMinterm(tabVar, 0, n-1, 0);
  Bdd_Fortify(s->table[0].p);

  /* MAKE BDDs FOR OTHER ACTIONS */

  for (i=1; i<s->numActions; i++) {
    s->table[i].p = EncodeMinterm(tabVar, 2*i, n-1, 1);
    Bdd_Fortify(s->table[i].p);
  }

  s->encoded = 1;
  PaSaveSortActions(s);
  return;
} 

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_EncodeProcess(Est_String name)
{
  Pa_Process *p;
  Bdd_Edge tau;
  Bdd_Edge var;
  Bdd_Edge d,e;
  Bdd_Edge states;
  Bdd_Edge *tabVarR;
  Bdd_Edge *tabVarS;
  Est_String x;
  int i,n,w;

  i = Pa_FindProcess(name);
  if (i == -1) {
    printf("Process does not exists.\n");
    return;
  }

  p = &pa_processTable[i];

  if (p->sort == -1) {
    printf("ERROR Pa_EncodeProcess: Sort is not valid!\n");
    return;
  }

  if (p->sort >= pa_sorts) {
    printf("ERROR Pa_EncodeProcess: Sort is not valid!\n");
    return;
  }

  /* po potrebi kodiramo sort */
  if (!pa_sortTable[p->sort].encoded) {
    /*
    printf("Encoding sort:\n");
    printf("  %s...",pa_sortTable[p->sort].name);
    */
    Pa_EncodeSort(pa_sortTable[p->sort].name);
    /*
    printf(" OK\n");
    */
  }

  /* poiscemo dogodek TAU - za kontorolo */
  tau = pa_sortTable[p->sort].table[0].p;

  /* dolocimo stevilo potrebnih spremenljivk za kodiranje stanj */
  if (p->numStates == 1) n = 1;
    else n = (int)(0.9999999+log(p->numStates)/log(2.0));
  p->numVar = n;

  /* tvorimo spremenljivke in dolocimo utezi */
  tabVarR = (Bdd_Edge *) malloc(n * sizeof(Bdd_Edge));
  tabVarS = (Bdd_Edge* ) malloc(n * sizeof(Bdd_Edge));
  x = (Est_String) malloc(127);
  w = 1;
  for (i=0; i<n; i++) {

    /* HERE IS GIVEN THE ORDER OF r,s,p,q VARIABLES */

    /* variable 'r' */
    sprintf(x,"r%d<%s>",i,name);  /* format "r%d%s\0" */
    var = Bdd_FoaTerminal(x);
    Pa_InitVariable(Bdd_GetVariable(var));
    pa_variableTable[Bdd_GetVariable(var)].weight = w;
    tabVarR[i] = var;

    /* variable 's' */
    sprintf(x,"s%d<%s>",i,name);  /* format "s%d%s\0" */
    var = Bdd_FoaTerminal(x);
    Pa_InitVariable(Bdd_GetVariable(var));
    pa_variableTable[Bdd_GetVariable(var)].weight = w;
    tabVarS[i] = var;

    /* variable 'p' */
    sprintf(x,"p%d<%s>",i,name);  /* format "p%d%s\0" */
    var = Bdd_FoaTerminal(x);
    Pa_InitVariable(Bdd_GetVariable(var));
    pa_variableTable[Bdd_GetVariable(var)].weight = w;

    /* variable 'q' */
    sprintf(x,"q%d<%s>",i,name);   /* format "q%d%s\0" */
    var = Bdd_FoaTerminal(x);
    Pa_InitVariable(Bdd_GetVariable(var));
    pa_variableTable[Bdd_GetVariable(var)].weight = w;

    w = w * 2;
  }

  free(x);

  /* zgradimo BDD za vsako stanje */
  for (i=0; i<p->numStates; i++) {
    p->tableS[i].bddR = EncodeMinterm(tabVarR, i, n-1, 0);
    p->tableS[i].bddS = EncodeMinterm(tabVarS, i, n-1, 0);
    Bdd_Fortify(p->tableS[i].bddR);
    Bdd_Fortify(p->tableS[i].bddS);
  }
  states = bdd_termFalse;
  for (i=0; i<p->numStates; i++) {
    states = Bdd_ITE(states,bdd_termTrue,p->tableS[i].bddR);
  }
  p->stateBDD = states;
  Bdd_Fortify(p->stateBDD);

  /* zgradimo prehajalno relacijo */
  d = bdd_termFalse;
  for (i=0; i<p->numTransitions; i++) {
    if (p->tableT[i].type) {
      e = Bdd_ITE(pa_sortTable[p->sort].a0,
              pa_sortTable[p->sort].table[p->tableT[i].action].p,
              bdd_termFalse);
    } else {
      e = Bdd_ITE(pa_sortTable[p->sort].a0,
              bdd_termFalse,
              pa_sortTable[p->sort].table[p->tableT[i].action].p);
    }

    e = Bdd_ITE(e,p->tableS[p->tableT[i].source].bddR,bdd_termFalse);
    e = Bdd_ITE(e,p->tableS[p->tableT[i].dest].bddS,bdd_termFalse);
    d = Bdd_ITE(d,bdd_termTrue,e);
  }
  Bdd_Fortify(d);
  p->d =d;

  /* zgradimo stab BDD */
  d = bdd_termTrue;
  for(i=0; i<p->numVar; i++) {
    e = Bdd_ITE(tabVarR[i],tabVarS[i],Bdd_NOT(tabVarS[i]));
    d = Bdd_ITE(d,e,bdd_termFalse);
  }
  Bdd_Fortify(d);
  p->stab =d;

  p->encoded = 1;
  PaSaveProcStates(p);
  Bdd_SaveFormula("D_",name,p->d);
  Bdd_SaveFormula("Initial_",p->name,p->tableS[p->initial].bddR);
  Bdd_SaveFormula("S_",p->name,p->stateBDD);

  /* POMOZNI IZPISI */

  /*
  Pa_DecodeProcessStates(p,p->stateBDD,FALSE);
  Pa_DecodeProcessTR(p,p->d,FALSE);
  Pa_DecodeProcessTR(p,Bdd_ITE(tau,bdd_termFalse,p->d),FALSE);
  printf("Stevilo BDD vozlisc = %d.\n",Bdd_NodeNumber(p->d));

  printf("\n");
  */
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_DecodeSortActions(Pa_Sort *s, Bdd_Edge F, Est_Boolean write)
{
  int i,maxAction;
  int number;

  if(s->encoded) {

    maxAction = 1;
    for (i=0; i<s->numVar; i++) maxAction= maxAction * 2;
    number = DecodeSortBddA(s, F, bdd_termTrue.mark, 0, maxAction-1, write);
    if (!write) printf("Stevilo dogodkov = %d.\n",number);

  } else {
    printf("Sort %s is not encoded.\n",s->name);
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_DecodeProcess(Est_String name, Est_Boolean write)
{
  Pa_Process *p;
  int i,maxAction,maxState;
  int number;

  i = Pa_FindProcess(name);
  if (i == -1) return;

  p = &pa_processTable[i];

  if(p->encoded) {
    printf("\nPROCESS %s\n",p->name); 
    printf("Initial state: %s\n\n",p->tableS[p->initial].name);

    maxAction = 1;
    for (i=0; i<pa_sortTable[p->sort].numVar; i++) maxAction = maxAction * 2;
    maxState = 1;
    for (i=0; i<p->numVar; i++) maxState= maxState * 2;

    number = DecodeProcessBddRAS(p, p->d, bdd_termTrue.mark, 0, maxState-1,
                                 0, maxAction-1, 0, maxState-1, write);

    if (!write) printf("Number of transitions = %d.\n",number);

  } else {
    printf("Process %s is not encoded.\n",p->name);
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeProcessStates(Pa_Process *p, Bdd_Edge F, Est_Boolean write)
{
  int i,maxState;
  int number;

  if(p->encoded) {

    maxState = 1;
    for (i=0; i<p->numVar; i++) maxState= maxState * 2;
    number = DecodeProcessBddR(p,F,bdd_termTrue.mark,0,maxState-1,write);

  } else {
    printf("Process %s is not encoded.\n",p->name);
    number = -1;
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeProcessTR(Pa_Process *p, Bdd_Edge F, Est_Boolean write)
{
  int i,maxAction,maxState;
  int number;

  if(p->encoded) {
    maxAction = 1;
    for (i=0; i<pa_sortTable[p->sort].numVar; i++) maxAction = maxAction * 2;
    maxState = 1;
    for (i=0; i<p->numVar; i++) maxState= maxState * 2;

    number = DecodeProcessBddRAS(p, F, bdd_termTrue.mark, 0, maxState-1,
                                 0, maxAction-1, 0, maxState-1, write);

  } else {
    printf("Process %s is not encoded.\n",p->name);
    number = -1;
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_DecodeComposition(Est_String name, Est_Boolean write)
{
  Pa_Composition *c;
  int *decodingTableR;
  int *decodingTableS;
  int i,j,maxAction,maxState;
  int number;

  i = Pa_FindComposition(name);
  if (i == -1) return;

  c = &pa_compositionTable[i];

  decodingTableR = (int *) calloc (2 * c->numProcesses, sizeof(int));
  decodingTableS = (int *) calloc (2 * c->numProcesses, sizeof(int));

  maxAction = 1;
  for (i=0; i<pa_sortTable[c->sort].numVar; i++) maxAction = maxAction * 2;

  for (i=0; i<c->numProcesses; i++) {
    maxState = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++)
      maxState = maxState * 2;
    decodingTableR[2*i] = 0;
    decodingTableR[2*i+1] = maxState-1;
    decodingTableS[2*i] = 0;
    decodingTableS[2*i+1] = maxState-1;
  }

  setproccomp(c);
  number = DecodeCompBddRAS(c,c->transitionBDD,bdd_termTrue.mark,0,
                            maxAction-1,decodingTableR,decodingTableS,write);

  free(decodingTableR);
  free(decodingTableS);

  if (!write) printf("Stevilo prehodov = %d.\n",number);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeCompStates(Pa_Composition *c, Bdd_Edge F, Est_Boolean write)
{
  int *decodingTableR;
  int i,j,maxState;
  int number;

  decodingTableR = (int *) calloc (2 * c->numProcesses, sizeof(int));

  for (i=0; i<c->numProcesses; i++) {
    maxState = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++)
      maxState = maxState * 2;
    decodingTableR[2*i] = 0;
    decodingTableR[2*i+1] = maxState-1;
  }

  setproccomp(c);
  number = DecodeCompBddR(c, F, bdd_termTrue.mark, decodingTableR, write);

  free(decodingTableR);

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeCompTR(Pa_Composition *c, Bdd_Edge F, Est_Boolean write)
{
  int *decodingTableR;
  int *decodingTableS;
  int i,j,maxAction,maxState;
  int number;

  decodingTableR = (int *) calloc (2 * c->numProcesses, sizeof(int));
  decodingTableS = (int *) calloc (2 * c->numProcesses, sizeof(int));

  maxAction = 1;
  for (i=0; i<pa_sortTable[c->sort].numVar; i++) maxAction = maxAction * 2;

  for (i=0; i<c->numProcesses; i++) {
    maxState = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++)
      maxState = maxState * 2;
    decodingTableR[2*i] = 0;
    decodingTableR[2*i+1] = maxState-1;
    decodingTableS[2*i] = 0;
    decodingTableS[2*i+1] = maxState-1;
  }

  setproccomp(c);
  number = DecodeCompBddRAS(c, F, bdd_termTrue.mark, 0, maxAction-1,
                            decodingTableR, decodingTableS, write);

  free(decodingTableR);
  free(decodingTableS);

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeProcProcPair(Pa_Process *p1, Pa_Process *p2, Bdd_Edge F,
                      Est_Boolean write)
{
  int i,maxState1,maxState2;
  int number;

  if ((p1->encoded) && (p2->encoded)) {
    maxState1 = 1;
    for (i=0; i<p1->numVar; i++) maxState1 = maxState1 * 2;
    maxState2 = 1;
    for (i=0; i<p2->numVar; i++) maxState2 = maxState2 * 2;

    number = DecodeProcProcPairBddRP(p1, p2, F, bdd_termTrue.mark,
                                     0, maxState1-1, 0, maxState2-1, write);

  } else {
    printf("Processes are not encoded.\n");
    number = -1;
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeCompProcPair(Pa_Composition *c, Pa_Process *p, Bdd_Edge F,
                      Est_Boolean write)
{
  int i,j,maxState1,maxState2;
  int *decodingTableR;
  int number;

  decodingTableR = (int *) calloc (2 * c->numProcesses, sizeof(int));

  if(p->encoded) {
    maxState1 = 1;
    for (i=0; i<p->numVar; i++) maxState1 = maxState1 * 2;

    for (i=0; i<c->numProcesses; i++) {
      maxState2 = 1;
      for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++)
        maxState2 = maxState2 * 2;
      decodingTableR[2*i] = 0;
      decodingTableR[2*i+1] = maxState2-1;
    }

    setproccomp(c);
    number = DecodeCompProcPairBddRP(c, p, F, bdd_termTrue.mark,
                                 decodingTableR, 0, maxState1-1, write);

  } else {
    printf("Process is not encoded.\n");
    number = -1;
  }

  free(decodingTableR);

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeCompCompPair(Pa_Composition *c1, Pa_Composition *c2, Bdd_Edge F,
                      Est_Boolean write)
{
  int i,j,maxState;
  int *decodingTableR,*decodingTableS;
  int number;

  decodingTableR = (int *) calloc (2 * c1->numProcesses, sizeof(int));
  decodingTableS = (int *) calloc (2 * c2->numProcesses, sizeof(int));

  for (i=0; i<c1->numProcesses; i++) {
    maxState = 1;
    for (j=0; j<pa_processTable[c1->tableP[i]].numVar; j++)
      maxState = maxState * 2;
    decodingTableR[2*i] = 0;
    decodingTableR[2*i+1] = maxState-1;
  }

  for (i=0; i<c2->numProcesses; i++) {
    maxState = 1;
    for (j=0; j<pa_processTable[c2->tableP[i]].numVar; j++)
      maxState = maxState * 2;
    decodingTableS[2*i] = 0;
    decodingTableS[2*i+1] = maxState-1;
  }

  setproccomp(c1);
  setproccomp(c2);
  number = DecodeCompCompPairBddRP(c1, c2, F, bdd_termTrue.mark,
                                   decodingTableR, decodingTableS, write);

  free(decodingTableR);
  free(decodingTableS);

  return number;
}


/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Edge
EncodeMinterm(Bdd_Edge *tabVar, int n, int i, int min)
{
  int w;

  if (i < min) return bdd_termTrue;
  w = getWeight(tabVar[i]);
  if (w <= n)
    return Bdd_ITE(tabVar[i],EncodeMinterm(tabVar,n-w,i-1,min),bdd_termFalse);
  else
    return Bdd_ITE(tabVar[i],bdd_termFalse,EncodeMinterm(tabVar,n,i-1,min));
}

/**Function****************************************************************
  Synopsis    []
  Description [Actions are encoded with variables a]
  SideEffe Est_Boolean writects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeSortBddA(Pa_Sort *sort, Bdd_Edge F, Est_Boolean mark,
               int minA, int maxA, Est_Boolean write)
{
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (Bdd_isTerminal(F)) {
    if (!mark) number = DecodeSortAction(sort,minA,maxA,write);
  } else {
    ch = Bdd_GetVariableChar(F);
    switch (ch) {
      case 'a':
        {
          number = number +
            DecodeSortBddA(sort,Bdd_GetElse(F),mark,minA,maxA-getWeight(F),write);
          number = number +
            DecodeSortBddA(sort,Bdd_GetThen(F),mark,minA+getWeight(F),maxA,write);
          break;
        }
    }
  }

  return number;
}


/**Function****************************************************************
  Synopsis    []
  Description [States are encoded with variables r]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcessBddR(Pa_Process *proc, Bdd_Edge F, Est_Boolean mark,
                  int minR, int maxR, Est_Boolean write)
{
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (Bdd_isTerminal(F)) {
    if (!mark) number = DecodeProcessState(proc,minR,maxR,write);
  } else {
    ch = Bdd_GetVariableChar(F);
    switch (ch) {
      case 'r':
        {
          number = number +
            DecodeProcessBddR(proc,Bdd_GetElse(F),mark,minR,maxR-getWeight(F),write);
          number = number +
            DecodeProcessBddR(proc,Bdd_GetThen(F),mark,minR+getWeight(F),maxR,write);
          break;
        }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description [Transition relation is encoded with variables r,a,s]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcessBddRAS(Pa_Process *proc, Bdd_Edge F, Est_Boolean mark,
                    int minR, int maxR, int minA, int maxA,
                    int minS, int maxS, Est_Boolean write)
{
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (Bdd_isTerminal(F)) {
    if (!mark)
      number = DecodeProcessTR(proc,minR,maxR,minA,maxA,minS,maxS,write);
  } else {
    ch = Bdd_GetVariableChar(F);
    switch (ch) {
      case 'r': 
        {
          number = number + DecodeProcessBddRAS(proc,Bdd_GetElse(F),mark,
                                      minR,maxR-getWeight(F),
                                      minA,maxA,minS,maxS,write);
          number = number + DecodeProcessBddRAS(proc,Bdd_GetThen(F),mark,
                                      minR+getWeight(F),maxR,
                                      minA,maxA,minS,maxS,write);
          break;
        }

      case 'a':
        {
          number = number + DecodeProcessBddRAS(proc,Bdd_GetElse(F),mark,
                                      minR,maxR,
                                      minA,maxA-getWeight(F),
                                      minS,maxS,write);
          number = number + DecodeProcessBddRAS(proc,Bdd_GetThen(F),mark,
                                      minR,maxR,
                                      minA+getWeight(F),maxA,
                                      minS,maxS,write);
          break;
        }

      case 's': 
        {
          number = number + DecodeProcessBddRAS(proc,Bdd_GetElse(F),mark,
                                      minR,maxR,minA,maxA,
                                      minS,maxS-getWeight(F),write);
          number = number + DecodeProcessBddRAS(proc,Bdd_GetThen(F),mark,
                                      minR,maxR,minA,maxA,
                                      minS+getWeight(F),maxS,write);
          break;
        }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description [States are encoded with variables r]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompBddR(Pa_Composition *comp, Bdd_Edge F, Est_Boolean mark,
               int *decodingTableR, Est_Boolean write)
{
  int i,old;
  char ch;
  Est_String pname;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (Bdd_isTerminal(F)) {
    if (!mark) number = DecodeCompState(comp,decodingTableR,write);
  } else {
    ch = Bdd_GetVariableChar(F);
    switch (ch) {
      case 'r': {
         	  i = getProcComp(F);

                  old = decodingTableR[2*i+1]; /* maxR */
                  decodingTableR[2*i+1] =
                    decodingTableR[2*i+1] - getWeight(F); /* maxR - weight */
                  number = number +
                    DecodeCompBddR(comp,Bdd_GetElse(F),mark,decodingTableR,write);
                  decodingTableR[2*i+1] = old;

                  old = decodingTableR[2*i]; /* minR */
                  decodingTableR[2*i] =
                    decodingTableR[2*i] + getWeight(F); /* minR + weight */
                  number = number +
                    DecodeCompBddR(comp,Bdd_GetThen(F),mark,decodingTableR,write);
                  decodingTableR[2*i] = old;

                  break;
                }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description [Transition relation is encoded with variables r,a,s.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompBddRAS(Pa_Composition *comp, Bdd_Edge F, Est_Boolean mark,
                    int minA, int maxA, int *decodingTableR,
                    int *decodingTableS, Est_Boolean write)
{
  int i,old;
  char ch;
  Est_String pname;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (Bdd_isTerminal(F)) {
    if (!mark) {
       number =
         DecodeCompTR(comp,minA,maxA,decodingTableR,decodingTableS,write);
    }
  } else {
    ch = Bdd_GetVariableChar(F);
    switch (ch) {
      case 'r': {
         	  i = getProcComp(F);

                  old = decodingTableR[2*i+1]; /* maxR */
                  decodingTableR[2*i+1] =
                    decodingTableR[2*i+1] - getWeight(F); /* maxR - weight */
                  number = number +
                    DecodeCompBddRAS(comp,Bdd_GetElse(F),mark,minA,maxA,
                                     decodingTableR,decodingTableS,write);
                  decodingTableR[2*i+1] = old;

                  old = decodingTableR[2*i]; /* minR */
                  decodingTableR[2*i] =
                    decodingTableR[2*i] + getWeight(F); /* minR + weight */
                  number = number +
                    DecodeCompBddRAS(comp,Bdd_GetThen(F),mark,minA,maxA,
                                     decodingTableR,decodingTableS,write);
                  decodingTableR[2*i] = old;

                  break;
                }
      case 's': {
         	  i = getProcComp(F);

                  old = decodingTableS[2*i+1]; /* maxR */
                  decodingTableS[2*i+1] =
                    decodingTableS[2*i+1] - getWeight(F); /* maxS - weight */
                  number = number +
                    DecodeCompBddRAS(comp,Bdd_GetElse(F),mark,minA,maxA,
                                     decodingTableR,decodingTableS,write);
                  decodingTableS[2*i+1] = old;

                  old = decodingTableS[2*i]; /* minS */
                  decodingTableS[2*i] =
                    decodingTableS[2*i] + getWeight(F); /* minS + weight */
                  number = number +
                    DecodeCompBddRAS(comp,Bdd_GetThen(F),mark,minA,maxA,
                                     decodingTableR,decodingTableS,write);
                  decodingTableS[2*i] = old;

                  break;
                }
      case 'a': {
                  number = number +
                    DecodeCompBddRAS(comp,Bdd_GetElse(F),mark,
                                     minA,maxA-getWeight(F),
                                     decodingTableR,decodingTableS,write);

                  number = number +
                    DecodeCompBddRAS(comp,Bdd_GetThen(F),mark,
                                     minA+getWeight(F),maxA,
                                     decodingTableR,decodingTableS,write);

                  break;
                }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcProcPairBddRP(Pa_Process *p1, Pa_Process *p2, Bdd_Edge F,
                        Est_Boolean mark, int minR, int maxR, int minP,
                        int maxP, Est_Boolean write)
{
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (Bdd_isTerminal(F)) {
    if (!mark) {
      number = DecodeProcProcPair(p1,p2,minR,maxR,minP,maxP,write);
    }
  } else {
    ch = Bdd_GetVariableChar(F);
    switch (ch) {
      case 'r': {
                  number = number +
                    DecodeProcProcPairBddRP(p1,p2,Bdd_GetElse(F),mark,
                                            minR,maxR-getWeight(F),
                                            minP,maxP,write);
                  number = number +
                    DecodeProcProcPairBddRP(p1,p2,Bdd_GetThen(F),mark,
                                            minR+getWeight(F),maxR,
                                            minP,maxP,write);
                  break;
                }
      case 'p': {
                  number = number +
                    DecodeProcProcPairBddRP(p1,p2,Bdd_GetElse(F),mark,
                                            minR,maxR,
                                            minP,maxP-getWeight(F),write);
                  number = number +
                    DecodeProcProcPairBddRP(p1,p2,Bdd_GetThen(F),mark,
                                            minR,maxR,
                                            minP+getWeight(F),maxP,write);
                  break;
                }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompProcPairBddRP(Pa_Composition *comp, Pa_Process *p, Bdd_Edge F,
                        Est_Boolean mark, int *decodingTableR, int minP,
                        int maxP, Est_Boolean write)
{
  Est_String pname;
  int i,old;
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (Bdd_isTerminal(F)) {
    if (!mark) {
      number = DecodeCompProcPair(comp,p,decodingTableR,minP,maxP,write);
    }
  } else {
    ch = Bdd_GetVariableChar(F);
    switch (ch) {
      case 'r': {
         	  i = getProcComp(F);

                  old = decodingTableR[2*i+1]; /* maxR */
                  decodingTableR[2*i+1] =
                    decodingTableR[2*i+1] - getWeight(F); /* maxR - weight */
                  number = number +
                    DecodeCompProcPairBddRP(comp,p,Bdd_GetElse(F),mark,
                                            decodingTableR,minP,maxP,write);
                  decodingTableR[2*i+1] = old;

                  old = decodingTableR[2*i]; /* minR */
                  decodingTableR[2*i] =
                    decodingTableR[2*i] + getWeight(F); /* minR + weight */
                  number = number +
                    DecodeCompProcPairBddRP(comp,p,Bdd_GetThen(F),mark,
                                            decodingTableR,minP,maxP,write);
                  decodingTableR[2*i] = old;

                  break;
                }
      case 'p': {
                  number = number +
                    DecodeCompProcPairBddRP(comp,p,Bdd_GetElse(F),mark,
                                            decodingTableR,
                                            minP,maxP-getWeight(F),write);
                  number = number +
                    DecodeCompProcPairBddRP(comp,p,Bdd_GetThen(F),mark,
                                            decodingTableR,
                                            minP+getWeight(F),maxP,write);
                  break;
                }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompCompPairBddRP(Pa_Composition *c1, Pa_Composition *c2, Bdd_Edge F,
                        Est_Boolean mark, int *decodingTableR,
                        int *decodingTableS, Est_Boolean write)
{
  Est_String pname;
  int i,old;
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (Bdd_isTerminal(F)) {
    if (!mark) {
      number = DecodeCompCompPair(c1,c2,decodingTableR,decodingTableS,write);
    }
  } else {
    ch = Bdd_GetVariableChar(F);
    switch (ch) {
      case 'r': {
         	  i = getProcComp(F);

                  old = decodingTableR[2*i+1]; /* maxR */
                  decodingTableR[2*i+1] =
                    decodingTableR[2*i+1] - getWeight(F); /* maxR - weight */
                  number = number +
                    DecodeCompCompPairBddRP(c1,c2,Bdd_GetElse(F),mark,decodingTableR,
                                            decodingTableS,write);
                  decodingTableR[2*i+1] = old;

                  old = decodingTableR[2*i]; /* minR */
                  decodingTableR[2*i] =
                    decodingTableR[2*i] + getWeight(F); /* minR + weight */
                  number = number +
                    DecodeCompCompPairBddRP(c1,c2,Bdd_GetThen(F),mark,decodingTableR,
                                            decodingTableS,write);
                  decodingTableR[2*i] = old;

                  break;
                }

      case 'p': {
         	  i = getProcComp(F);

                  old = decodingTableS[2*i+1]; /* maxP */
                  decodingTableS[2*i+1] =
                    decodingTableS[2*i+1] - getWeight(F); /* maxP - weight */
                  number = number +
                    DecodeCompCompPairBddRP(c1,c2,Bdd_GetElse(F),mark,decodingTableR,
                                            decodingTableS,write);
                  decodingTableS[2*i+1] = old;

                  old = decodingTableS[2*i]; /* minP */
                  decodingTableS[2*i] =
                    decodingTableS[2*i] + getWeight(F); /* minP + weight */
                  number = number +
                    DecodeCompCompPairBddRP(c1,c2,Bdd_GetThen(F),mark,decodingTableR,
                                            decodingTableS,write);
                  decodingTableS[2*i] = old;

                  break;
                }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeSortAction(Pa_Sort *s, int minA, int maxA, Est_Boolean write)
{
  int i,fA;
  int number;

  fA = 1;
  for (i=1; i<s->numVar; i++) fA = fA * 2;

  if (write) printf("<");
  number = DecodeAction(s,minA,maxA,fA,write);
  if (write) printf(">");

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcessState(Pa_Process *p, int minR, int maxR, Est_Boolean write)
{
  int i,fR;
  int number;

  fR = 1;
  for (i=1; i<p->numVar; i++) fR = fR * 2;

  if (write) printf("<");
  number = DecodeState(p,minR,maxR,fR,write);
  if (write) printf(">");

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcessTR(Pa_Process *p, int minR, int maxR,
                int minA, int maxA, int minS, int maxS, Est_Boolean write)
{
  int i,fA,fS;
  int nr,na,ns,number;

  fS = 1;
  fA = 1;

  for (i=1; i<p->numVar; i++) fS = fS * 2;
  for (i=1; i<pa_sortTable[p->sort].numVar; i++) fA = fA * 2;

  if (write) printf("(");
  nr = DecodeState(p,minR,maxR,fS,write);
  if (write) printf(") --- ");
  na = DecodeAction(&pa_sortTable[p->sort],minA,maxA,fA,write);
  if (write) printf(" ---> (");
  ns = DecodeState(p,minS,maxS,fS,write);
  if (write) printf(")\n");

  number = nr * na * ns;
  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompState(Pa_Composition *c, int *decodingTableR, Est_Boolean write)
{
  int i,j,fS;
  int number;

  number = 1;
  if (write) printf("<");

  for (i=0; i<c->numProcesses; i++) {
    fS = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++) fS = fS * 2;
    if (write) printf("(");
    number = number * DecodeState(&pa_processTable[c->tableP[i]],
                      decodingTableR[2*i],decodingTableR[2*i+1],fS,write);
    if (write) printf("),");
  }

  if (write) printf(">\n");
  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompTR(Pa_Composition *c, int minA, int maxA,
             int *decodingTableR, int *decodingTableS, Est_Boolean write)
{
  int i,j,fA,fS;
  int nr,na,ns,number;

  nr = 1;
  ns = 1;
  if (write) printf("(");

  for (i=0; i<c->numProcesses; i++) {
    fS = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++) fS = fS * 2;
    if (write) printf("(");
    nr = nr * DecodeState(&pa_processTable[c->tableP[i]],
                decodingTableR[2*i],decodingTableR[2*i+1],fS,write);
    if (write) printf("),");
  }

  if (write) printf(") --- ");

  fA = 1;
  for (i=1; i<pa_sortTable[c->sort].numVar; i++) fA = fA * 2;
  na = DecodeAction(&pa_sortTable[c->sort],minA,maxA,fA,write);

  if (write) printf(" ---> (");

  for (i=0; i<c->numProcesses; i++) {
    fS = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++) fS = fS * 2;
    if (write) printf("(");
    ns = ns * DecodeState(&pa_processTable[c->tableP[i]],
                decodingTableS[2*i],decodingTableS[2*i+1],fS,write);
    if (write) printf("),");
  }

  if (write) printf(")\n");

  number = nr * na * ns;
  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcProcPair(Pa_Process *p1, Pa_Process *p2, int minR, int maxR,
                   int minP, int maxP, Est_Boolean write)
{
  int i, fR, fP;
  int number;

  fR = 1;
  fP = 1;

  for (i=1; i<p1->numVar; i++) fR = fR * 2;
  for (i=1; i<p2->numVar; i++) fP = fP * 2;

  number = 1;
  if (write) printf("<(");
  number = number * DecodeState(p1,minR,maxR,fR,write);
  if (write) printf("),(");
  number = number * DecodeState(p2,minP,maxP,fP,write);
  if (write) printf(")>\n");

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompProcPair(Pa_Composition *c, Pa_Process *p, int *decodingTableR,
                   int minP, int maxP, Est_Boolean write)
{
  int i,j,fS,fP;
  int nr,ns,number;

  nr = 1;
  if (write) printf("<<");

  for (i=0; i<c->numProcesses; i++) {
    fS = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++) fS = fS * 2;
    if (write) printf("(");
    nr = nr * DecodeState(&pa_processTable[c->tableP[i]],
                decodingTableR[2*i],decodingTableR[2*i+1],fS,write);
    if (write) printf("),");
  }

  if (write) printf(">,(");

  fP = 1;
  for (i=1; i<p->numVar; i++) fP = fP * 2;
  ns = DecodeState(p,minP,maxP,fP,write);

  if (write) printf(")>\n");

  number = nr * ns;
  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompCompPair(Pa_Composition *c1, Pa_Composition *c2,
                   int *decodingTableR, int *decodingTableS,
                   Est_Boolean write)
{
  int i,j,fR,fS;
  int nr,ns,number;

  nr = 1;
  ns = 1;
  if (write) printf("<<");

  for (i=0; i<c1->numProcesses; i++) {
    fR = 1;
    for (j=0; j<pa_processTable[c1->tableP[i]].numVar; j++) fR = fR * 2;
    if (write) printf("(");
    nr = nr * DecodeState(&pa_processTable[c1->tableP[i]],
                decodingTableR[2*i],decodingTableR[2*i+1],fR,write);
    if (write) printf("),");
  }

  if (write) printf(">,<");

  for (i=0; i<c2->numProcesses; i++) {
    fS = 1;
    for (j=0; j<pa_processTable[c2->tableP[i]].numVar; j++) fS = fS * 2;
    if (write) printf("(");
    ns = ns * DecodeState(&pa_processTable[c2->tableP[i]],
                decodingTableS[2*i],decodingTableS[2*i+1],fS,write);
    if (write) printf("),");
  }

  if (write) printf(">>\n");

  number = nr * ns;
  return number;
}


/**Function****************************************************************
  Synopsis    []
  Description [Type variable (a0) should have weight = 1.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeAction(Pa_Sort *s, int min, int max, int f, Est_Boolean write)
{
  int number;

  if (min > max) {
    printf("DecodeAction ERROR!\n");
    printf("name = %s, min = %d, max = %d\n",s->name,min,max);
    exit(1);
  }

  number = 0;
  if (min == max) {

    if (min >= 2 * s->numActions) {
      if (write) {
        if (min % 2) printf("!!,"); else printf("??,");
      }
    } else {
      if (write) {
        if (min > 0) {
          printf(s->table[(min)/2].name);
          if (min % 2) printf("!, "); else printf("?, ");
        } else {
          printf("TAU, ");
        }
      }
      number = 1;
    }

  } else {

    if ( (max - min) >= f ) {
      number = number + DecodeAction(s,min+f,max,f/2,write);
      number = number + DecodeAction(s,min,max-f,f/2,write);
    } else {
      number = number + DecodeAction(s,min,max,f/2,write);
    }

  }
  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeState(Pa_Process *p, int min, int max, int f, Est_Boolean write)
{
  int number;

  if (min > max) {
    printf("DecodeState ERROR!\n");
    printf("name = %s, min = %d, max = %d\n",p->name,min,max);
    exit(1);
  }

  number = 0;
  if (min == max) {

    if (min >= p->numStates) {
      if (write) printf("??,");
    } else {
      if (write) {
        printf(p->tableS[min].name);
        printf(",");
      }
      number = 1;
    }

  } else {

    if ( (max - min) >= f ) {
      number = number + DecodeState(p,min+f,max,f/2,write);
      number = number + DecodeState(p,min,max-f,f/2,write);
    } else {
      number = number + DecodeState(p,min,max,f/2,write);
    }

  }
  return number;
}

/**Function****************************************************************
  Synopsis    [Function setproccomp]
  Description [Set proccomp (to which proccess it belongs) for variables.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
setproccomp(Pa_Composition *comp)
{
  int i,n;
  Est_String name,pname;

  for (i=0;i<pa_variables;i++) {
    name = Bdd_GetIthVariableName(i);
    n = strspn(&name[1],"0123456789");
    pname = strdup(&name[n+2]);
    pname[strlen(pname)-1] = 0;
    pa_variableTable[i].proccomp = Pa_FindProcessComp(comp,pname);
    free(pname);
  }
}

/**Function****************************************************************
  Synopsis    [Function getWeight]
  Description [Return weight of top variable.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
getWeight(Bdd_Edge f)
{
  return pa_variableTable[Bdd_GetVariable(f)].weight;
}

/**Function****************************************************************
  Synopsis    [Function getProcComp]
  Description [Return proccomp of top variable.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
getProcComp(Bdd_Edge f)
{
  return pa_variableTable[Bdd_GetVariable(f)].proccomp;
}
