/**CFile*******************************************************************
  PackageName [pa]
  Synopsis    [Package 'pa' is a framework for implementation of
               process algebrae similar to Milner's CCS and Hoare's CSP.]

  FileName    [paMain.c]
  Revision    [$Revision: 13 $]
  Date        [$Date: 2007-07-19 14:02:50 +0200 (Thu, 19 Jul 2007) $]
  Authors     [Robert Meolic (meolic@uni-mb.si)]
  Description [File paMain.c contains main functions.]
  SeeAlso     [pa.h, paInt.h]

  Copyright   [This file is part of EST (Efficient Symbolic Tools).
               Copyright (C) 1999, 2003, 2006, 2007
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               EST is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               EST is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

#include "paInt.h"

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

int pa_status = 0;                     /* initialization status */
int pa_variables;                      /* number of variables */
int pa_sorts;                          /* number of sorts */
int pa_processes;                      /* number of processes */
int pa_compositions;                   /* number of compositions */
Pa_Variable *pa_variableTable;         /* variable table */
Pa_Sort *pa_sortTable;                 /* table of sorts */
Pa_Process *pa_processTable;           /* table of processes */
Pa_Composition *pa_compositionTable;   /* table of compositions */

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static int wordNumber(Est_String s);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

#ifndef TCLOUT
   void Pa_Init(void) {}
#endif

/**Function****************************************************************
  Synopsis    [Function Pa_InitPkg.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_InitPkg()
{
  if (pa_status == 1) {
    printf("\nProcess algebra package is already initialized. ");
    return;
  } else {
    pa_status = 1;
  }

  pa_variables = 0;
  pa_sorts = 0;
  pa_processes = 0;
  pa_compositions = 0;

  /* INITIALIZATION OF VARIABLE TABLE */
  /* BDD VARIABLE WITH INDEX 0 IS CONSTANT 1 */
  if (!(pa_variableTable = (Pa_Variable *) malloc(sizeof(Pa_Variable)))) {
      fprintf(stderr,"\nPa_InitPkg: Out of memory error!\n");
      exit(1);
  }
  pa_variableTable[0].weight = -1;
  pa_variableTable[0].proccomp = -1;
}

/**Function****************************************************************
  Synopsis    [Function Pa_ExitPkg.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_ExitPkg()
{
}

/**Function****************************************************************
  Synopsis    [Function Pa_AboutPkg]
  Description [Pa_AboutPkg reports release of Process algebra package.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_AboutPkg() {
  printf("EST ");
  printf(EDITION);
  printf(", version ");
  printf(VERSION);
  printf(", ");
  printf("Process algebra package");
  printf("\n");
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_InitVariable(int i)
{

  /* VARIABLE TABLE IS NEVER EMPTY. AT LEAST, THERE IS ELEMENT '1' AT INDEX [0] */
  if (i > pa_variables) {
    pa_variables = i;
    if (!(pa_variableTable = (Pa_Variable *)
        realloc(pa_variableTable, (pa_variables+1) * sizeof(Pa_Variable)))) {
      fprintf(stderr,"\nPa_InitVariable: Out of memory error!\n");
      exit(1);
    }
  }

  pa_variableTable[i].weight = -1;    /*  weight = 1,2,4... (decoding )  */
  pa_variableTable[i].proccomp = -1;   /* process-composition (decoding) */
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_AddNewSort(Est_String name)
{
  Pa_Sort *s;
  int i = 0;
  int sortnum;
  Est_Boolean OK;

  if (pa_sorts == 0) {
    pa_sortTable = (Pa_Sort *) malloc (sizeof(Pa_Sort));
    s = pa_sortTable;
    pa_sorts = 1;
    sortnum = 0;
  } else {
    OK = FALSE;
    while (!OK && i<pa_sorts) {
      if (!strcmp(name,pa_sortTable[i].name)) {
        OK = TRUE;
      } else {
        i++;
      }
    }
    if (OK) {
      sortnum = i;
      s = &pa_sortTable[i];
      printf("    Warning: Sort %s overwritten.\n",name);
    } else {
      sortnum = pa_sorts;
      pa_sorts++;
      pa_sortTable = (Pa_Sort *) realloc (pa_sortTable,
                                  pa_sorts * sizeof(Pa_Sort));
      s = &pa_sortTable[pa_sorts-1];
    }
  }

  s->name = strdup(name);
  s->encoded = 0;

  /* ADD ACTION 'TAU' */

  s->numActions = 1;
  s->table = (Pa_Action *) malloc (sizeof(Pa_Action));
  s->table[0].name = strdup("TAU");
  s->table[0].p = bdd_termNull;

  return (sortnum);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindSort(Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindSort got null string!\n");
    return -1;
  }
  while (i<pa_sorts) {
    if (!strcmp(name,pa_sortTable[i].name))
      return i;
    else
      i++;
  }

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FOASortAction(Pa_Sort *sort, Est_String name)
{
  Pa_Action *a;
  int i = 0;

  if (!sort) return -1;

  if (!name) {
    printf("\nError: Pa_FOASortAction got null string!\n");
    return -1;
  }

  /* SORT CANNOT BE EMPTY. ALWAYS EXIST AT LEAST ACTION 'TAU' */

  while (i<sort->numActions) {
    if (!strcmp(name,sort->table[i].name))
      return i;
    else
      i++;
  }

  (sort->numActions)++;
  sort->table = (Pa_Action *) realloc (sort->table,
                                 sort->numActions * sizeof(Pa_Action));
  a = &sort->table[sort->numActions-1];

  a->name = strdup(name);
  a->p = bdd_termNull;

  return (sort->numActions-1);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindSortAction(Pa_Sort *sort, Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindSortAction got null string!\n");
    return -1;
  }
  if (!sort) return -1;
  while (i<sort->numActions) {
    if (!strcmp(name,sort->table[i].name))
      return i;
    else
      i++;
  }

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_AddNewProcess(Est_String name)
{
  Pa_Process *p;
  Est_String init;
  int i = 0;
  int procnum;
  Est_Boolean OK;

  if (pa_processes == 0) {
    pa_processTable = (Pa_Process *) malloc (sizeof(Pa_Process));
    p = pa_processTable;
    pa_processes = 1;
    procnum = 0;
  } else {
    OK = FALSE;
    while (!OK && i<pa_processes) {
      if (!strcmp(name,pa_processTable[i].name)) {
        OK = TRUE;
      } else {
        i++;
      }
    }
    if (OK) {
      procnum = i;
      p = &pa_processTable[i];
      printf("    Warning: Process %s overwritten.\n",name);
    } else {
      procnum = pa_processes;
      pa_processes++;
      pa_processTable = (Pa_Process *) realloc (
                        pa_processTable, pa_processes * sizeof(Pa_Process));
      p = &pa_processTable[pa_processes-1];
    }
  }

  p->name = strdup(name);
  p->encoded = 0;
  p->sort = -1;
  p->numStates = 0;
  p->tableS = NULL;
  p->initial = 0;
  p->numTransitions = 0;
  p->tableT = NULL;
  p->numVar = 0;
  p->d = bdd_termNull;
  p->stab = bdd_termNull;

  return (procnum);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindProcess(Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindProcess got null string!\n");
    return -1;
  }
  while (i<pa_processes) {
    if (!strcmp(name,pa_processTable[i].name)) {
      return i;
    } else {
      i++;
    }
  }

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FOAStateProcess(Pa_Process *proc, Est_String name)
{
  Pa_State *s;
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FOAStateProcess got null string!\n");
    return -1;
  }

  if (!proc) return -1;
  if (proc->numStates == 0) {
    proc->numStates = 1;
    proc->tableS = (Pa_State *) malloc (sizeof(Pa_State));
    s = proc->tableS;
  } else {
    while (i<proc->numStates) {
      if (!strcmp(name,proc->tableS[i].name)) {
        return i;
      } else {
        i++;
      }
    }

    (proc->numStates)++;
    proc->tableS = (Pa_State *) realloc (proc->tableS,
                                   proc->numStates * sizeof(Pa_State));
    s = &proc->tableS[proc->numStates-1];
  }    

  s->name = strdup(name);
  s->bddR = bdd_termNull;
  s->bddS = bdd_termNull;
  s->acc = bdd_termNull;
  s->minacc = bdd_termNull;

  return (proc->numStates-1);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindStateProcess(Pa_Process *proc, Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindStateProcess got null string!\n");
    return -1;
  }

  if (!proc) return -1;
  while (i<proc->numStates) {
    if (!strcmp(name,proc->tableS[i].name)) {
      return i;
    } else {
      i++;
    }
  }

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description [Add a state to the proces. The name of the state will be
               composed of given prefix and the node number.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FOANextStateProcess(Pa_Process *proc, Est_String prefix)
{
  int i;
  Est_String x,y;

  if (!prefix) {
    printf("\nError: Pa_FOANextStateProcess got null string!\n");
    return -1;
  }

  x = (Est_String) malloc(10);
  sprintf(x,"%d",proc->numStates);   /* format "%d\0" !!! */

  y = strdup(prefix);
  y = (char*) realloc(y,strlen(y) + strlen(x) + 1);
  strcat(y,x);

  i = Pa_FOAStateProcess(proc,y);
  return i;
}

/**Function****************************************************************
  Synopsis    []
  Description [type 0 = output, type 1 = input]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FOATransition(Pa_Process *proc, int state1, int action, int type,
                 int state2)
{
  Pa_Transition *t;
  int i = 0;

  if (!proc) return -1;
  if (proc->numTransitions == 0) {
    proc->numTransitions = 1;
    proc->tableT = (Pa_Transition *) malloc (sizeof(Pa_Transition));
    t = proc->tableT;
  } else {
    while (i<proc->numTransitions) {
      if ((state1 == proc->tableT[i].source) &&
          (action == proc->tableT[i].action) &&
          (type == proc->tableT[i].type) &&
          (state2 == proc->tableT[i].dest))
      {
        return i;
      } else {
        i++;
      }
    }

    (proc->numTransitions)++;
    proc->tableT = (Pa_Transition *) realloc (proc->tableT, 
                                proc->numTransitions * sizeof(Pa_Transition));
    t = &proc->tableT[proc->numTransitions-1];
  }

  t->source = state1;
  t->action = action;
  t->type = type;
  t->dest = state2;

  return (proc->numTransitions-1);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [Renamed process has the same seq as the original one.]
  SeeAlso     []
  ************************************************************************/

void
Pa_RenameProcess(Est_String old, Est_String new)
{
  int oldproc,newproc;
  int i,j;
  Pa_Process *p;
  Est_String state;
  Est_String sup1, sup2;

  oldproc = Pa_FindProcess(old);
  if (oldproc == -1) {
    /*
    printf("Process does not exists.\n");
    */
    return;
  }

  newproc = Pa_FindProcess(new);
  if (newproc != -1) {
    printf(" (process overwritten) ");
    Pa_DeleteProcess(newproc);
  }

  /* CHANGE NAME OF PROCESS */
  free(pa_processTable[oldproc].name);
  pa_processTable[oldproc].name = strdup(new);

  /* CHANGE NAME OF STATES */
  p = &pa_processTable[oldproc];
  for (i=0; i<p->numStates; i++) {
    state = strdup(p->tableS[i].name);
    if (strcmp(state,"")) {

      /* STATES WILL GET THE NAME OF NEW PROCESS */
      j=strcspn(state,"<");
      sup1 = (Est_String) malloc(j+1);
      memcpy(sup1,state,j);
      sup1[j]=0;
      sup2 = (Est_String) malloc(65535);
      sprintf(sup2,"%s<%s>",sup1,new);
      free(p->tableS[i].name);
      p->tableS[i].name = strdup(sup2);
      free(sup1);
      free(sup2);
      free(state);

    }
  }

  /* RE-ENOCDE TO GET CORRRECT NAMES FOR BDD VARIABLES */
  if (p->encoded) {
    Pa_EncodeProcess(p->name);
  }

}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [Process is erased and not removed from the table!
               The sequence numbers of other processes are not changed.]
  SeeAlso     []
  ************************************************************************/

void
Pa_DeleteProcess(int num)
{
  int i;
  Pa_Process *p;

  p = &pa_processTable[num];

  if (p->name) free(p->name);
  p->name = strdup("DELETED");

  for(i=0; i<p->numStates; i++) {
    if (p->tableS[i].name) {
      free(p->tableS[i].name);
      p->tableS[i].name = NULL;
    }
  }

  free(p->tableS);
  free(p->tableT);

  p->encoded = 0;
  p->sort = -1;
  p->numStates = 0;
  p->tableS = NULL;
  p->initial = -1;
  p->numTransitions = 0;
  p->tableT = NULL;
  p->numVar = 0;
  p->d = bdd_termNull;
  p->stab = bdd_termNull;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_AddNewComposition(Est_String name, int type, int numP, int numA)
{
  Pa_Composition *c;
  int i = 0;
  int compnum;
  Est_Boolean OK;

  if (pa_compositions == 0) {
    pa_compositionTable = (Pa_Composition *) malloc (sizeof(Pa_Composition));
    c = pa_compositionTable;
    pa_compositions = 1;
    compnum = 0;
  } else {
    OK = FALSE;
    while (!OK && i<pa_compositions) {
      if (!strcmp(name,pa_compositionTable[i].name)) {
        OK = TRUE;
      } else {
        i++;
      }
    }
    if (OK) {
      compnum = i;
      c = &pa_compositionTable[i];
    } else {
      compnum = pa_compositions;
      pa_compositions++;
      pa_compositionTable = (Pa_Composition *) realloc (pa_compositionTable,
                            pa_compositions * sizeof(Pa_Composition));
      c = &pa_compositionTable[pa_compositions-1];
    }
  }

  c->name = strdup(name);
  c->type = type;
  c->numProcesses = numP;
  c->tableP = (int *) calloc (numP,sizeof(int));
  c->numActions = numA;
  c->tableA = (int *) calloc (numA,sizeof(int));
  c->sort = -1;
  c->initialBDD = bdd_termNull;
  c->stateBDD = bdd_termNull;
  c->transitionBDD = bdd_termNull;
  c->stab = bdd_termNull;
  c->di = (Bdd_Edge *) calloc (numP,sizeof(Bdd_Edge));
  c->dij = (Bdd_Edge *) calloc (numP*numP,sizeof(Bdd_Edge));

  return (compnum);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_RemoveComposition(int num)
{
  Pa_Composition *c;

  c = &pa_compositionTable[num];
  free(c->name);
  free(c->tableP);
  free(c->tableA);
  free(c->di);
  free(c->dij);

  if (pa_compositions == 1) {
    free(pa_compositionTable);
    pa_compositions = 0;
  } else {
    pa_compositions--;
    if (num != pa_compositions) {
      pa_compositionTable[num] = pa_compositionTable[pa_compositions];
    }
    pa_compositionTable = (Pa_Composition *) realloc (pa_compositionTable,
                            pa_compositions * sizeof(Pa_Composition));
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindComposition(Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindComposition got null string!\n");
    return -1;
  }

  while (i<pa_compositions) {
    if (!strcmp(name,pa_compositionTable[i].name)) {
      return i;
    } else {
      i++;
    }
  }

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindProcessComp(Pa_Composition *c, Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindProcessComp got null string!\n");
    return -1;
  }

  if (!c) return -1;
  while (i<c->numProcesses) {
    if (!strcmp(name,pa_processTable[c->tableP[i]].name))
      return i;
    else
      i++;
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
PaSaveSortActions(Pa_Sort *s)
{
  int i;
  Est_String x;

  for (i=1; i<s->numActions; i++) {
    x = strdup("Action_");
    x = (char*) realloc(x,10 + strlen(s->name) +
                          strlen(s->table[i].name));
    strcat(x,s->name);
    strcat(x,"_");
    strcat(x,s->table[i].name);
    strcat(x,"!");
    Bdd_AddFormula(x,Bdd_ITE(s->a0,s->table[i].p,bdd_termFalse));
    x[strlen(x)-1] = '?';
    Bdd_AddFormula(x,Bdd_ITE(s->a0,bdd_termFalse,s->table[i].p));
    free(x);
  }

  /* ACTION TAU */

  x = strdup("Action_");
  x = (char*) realloc(x,12 + strlen(s->name));
  strcat(x,s->name);
  strcat(x,"_");
  strcat(x,"TAU");
  Bdd_AddFormula(x,s->table[0].p);
  free(x);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
PaSaveProcStates(Pa_Process *p)
{
  int i;
  Est_String x;

  for (i=0; i<(p->numStates); i++) {
    x = strdup("State_");
    x = (char*) realloc(x,8 + strlen(p->name) +
                          strlen(p->tableS[i].name));
    strcat(x,p->name);
    strcat(x,"_");
    strcat(x,p->tableS[i].name);
    Bdd_AddFormula(x,p->tableS[i].bddR);
    free(x);
  }
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
wordNumber(Est_String s)
{
  int num;
  int token;
  Est_String sup,sup_next;

  num = 0;
  sup = strdup(s);

  token = strspn(sup," ");
  while (strspn(sup," ") < strlen(sup)) {
    sup_next = (Est_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    num++;
    token = strcspn(sup," ");
    sup_next = (Est_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    token = strspn(sup," ");
  }

  free(sup);
  return num;
}
