/**CFile*******************************************************************
  PackageName [versis]
  Synopsis    [Package 'versis' provides basic tools for formal
               verification of concurrent systems]

  FileName    [versisEqv.c]
  Revision    [$Revision: 13 $]
  Date        [$Date: 2007-07-19 14:02:50 +0200 (Thu, 19 Jul 2007) $]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Mirjam Sepesy Maucec (mirjam.sepesy@uni-mb.si),
               Tatjana Kapus (kapus@uni-mb.si)]
  Description [File versisEqv.c provides different equivalence checking
               algorithms.]
  SeeAlso     [versis.h, versisInt.h]

  Copyright   [This file is part of EST (Efficient Symbolic Tools).
               Copyright (C) 1999, 2003, 2006, 2007
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               EST is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               EST is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

#include "versisInt.h"

/*-----------------------------------------------------------------------*/
/* Structure declarations                                                */
/*-----------------------------------------------------------------------*/

typedef struct {
  Bdd_Edge S;
  Est_Boolean b;
  int t;
} StateTable;

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

int STnum;
StateTable *ST;

int pom_n1;
int pom_n2;

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static Est_Boolean Bisimulation(Bdd_Edge init, Bdd_Edge eqs,
                                Bdd_Edge St1, Bdd_Edge St2,
                                Bdd_Edge D1, Bdd_Edge D2);

static Bdd_Edge Divergent(Bdd_Edge D, Bdd_Edge tau);

static Bdd_Edge BuildAcc(int par, Est_String name, Est_String newname,
                     Bdd_Edge D, Bdd_Edge init,
                     Bdd_Edge Ptau, Bdd_Edge trcl, Bdd_Edge Div, int s,
                     Bdd_Edge tau, Bdd_Edge *Acc, Bdd_Edge *AccFalse,
                     Bdd_Edge *AccNull, Bdd_Edge *InitNew);

static int BuildAccR(int par, Est_String name,
                     Bdd_Edge D, Bdd_Edge Ptau, Bdd_Edge S, Bdd_Edge tau, 
                     Bdd_Edge Div, Est_Boolean b, int s);

static Bdd_Edge MakeAcc(Bdd_Edge D, Bdd_Edge tau, Bdd_Edge S);

static void MinAcc(int numNS, Pa_State *tabNS, int numS, Pa_State *tabS);

static Est_Boolean CmpAcc(Bdd_Edge acc1, Bdd_Edge acc2);

static void AddStateTable(Bdd_Edge S, Est_Boolean b, int t);

static Est_Boolean FindStateTable(Bdd_Edge S, Est_Boolean b, int *t);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function Versis_Equivalence]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Est_Boolean
Versis_Equivalence(int type, int par1, Est_String name1,
                             int par2, Est_String name2)
{

  int n1,n2;
  Est_String pom;
  Est_Boolean r;

  r = FALSE;

  /* FIRST PARAMETER */

  if (par1 == 0) {
    n1 = Pa_FindProcess(name1);
    if (n1 == -1) {
      printf(" Process %s does not exist.\n",name1);
      return;
    }
    if (!pa_processTable[n1].encoded) {
      printf(" Process %s is not encoded.\n",name1);
      return;
    }
  }
  if (par1 == 1) {
    n1 = Pa_FindComposition(name1);
    if (n1 == -1) {
      printf(" Composition %s does not exist.\n",name1);
      return;
    }
  }

  /* SECOND PARAMETER */

  if (par2 == 0) {
    n2 = Pa_FindProcess(name2);
    if (n2 == -1) {
      printf(" Process %s does not exist.\n",name2);
      return;
    }
    if (!pa_processTable[n2].encoded) {
      printf(" Process %s is not encoded.\n",name2);
      return;
    }
  }
  if (par2 == 1) {
    n2 = Pa_FindComposition(name2);
    if (n2 == -1) {
      printf(" Composition %s does not exist.\n",name2);
      return;
    }
  }

  switch (type) {
    case 0: {
              r = VersisStrongEquivalence(par1,n1,par2,n2);
              break;
            }
    case 1: {
              r = VersisWeakEquivalence(par1,n1,par2,n2);
              break;
            }
    case 2: {
              r = VersisTestEquivalence(par1,n1,par2,n2);
              break;
            }
  }

  return r;
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description [par1, par2 = 0 (process) or 1 (composition)]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Est_Boolean
VersisStrongEquivalence(int par1, int n1, int par2, int n2)
{
  Bdd_Edge St1, St2, D1, D2;
  Bdd_Edge ini,ini1,ini2;
  Est_Boolean EQV;
  Est_String pom;

  if (par1 == 0) {
    D1 = pa_processTable[n1].d;
    St1 = pa_processTable[n1].stateBDD;
    ini1 = pa_processTable[n1].tableS[pa_processTable[n1].initial].bddR;
  } else {
    D1 = pa_compositionTable[n1].transitionBDD;
    St1 = pa_compositionTable[n1].stateBDD;
    ini1 = pa_compositionTable[n1].initialBDD;
  }

  if (par2 == 0) {
    D2 = pa_processTable[n2].d;
    St2 = pa_processTable[n2].stateBDD;
    ini2 = pa_processTable[n2].tableS[pa_processTable[n2].initial].bddR;
  } else {
    D2 = pa_compositionTable[n2].transitionBDD;
    St2 = pa_compositionTable[n2].stateBDD;
    ini2 = pa_compositionTable[n2].initialBDD;
  }

  /* produkt zacetnih stanj */
  ini = Bdd_ITE(ini1,Bdd_R2P(ini2),bdd_termFalse);
  Bdd_Fortify(ini);

  /* stanja D2 kodiramo s pomoznimi spremenljivkami */

  D2 = Bdd_RelOp(D2,bdd_termFalse,"R2P S2Q",TRUE);
  St2 = Bdd_R2P(St2);
  Bdd_Fortify(D2);
  Bdd_Fortify(St2);

  EQV = Bisimulation(ini,bdd_termTrue,St1,St2,D1,D2);

  return EQV;
}

/**Function****************************************************************
  Synopsis    []
  Description [par1, par2 = 0 (process) or 1 (composition)]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Est_Boolean
VersisWeakEquivalence(int par1, int n1, int par2, int n2)
{  
  Bdd_Edge St1, St2, D1, D2;
  Bdd_Edge tau, stab1, stab2;
  Bdd_Edge eps, eps1, eps2, eps_rp, eps_qs;
  Bdd_Edge tr1, tr2, tr1_novi, tr2_novi;
  Bdd_Edge ini,ini1,ini2;
  Bdd_Edge pom, pom1, pom2, pom3, eqs;
  Est_Boolean EQV;
  Est_String name1,name2;

  pom_n1 = n1;
  pom_n2 = n2;

  if (par1 == 0) {
    D1 = pa_processTable[n1].d;
    St1 = pa_processTable[n1].stateBDD;
    tau = pa_sortTable[pa_processTable[n1].sort].table[0].p;
    stab1 = pa_processTable[n1].stab;
    ini1 = pa_processTable[n1].tableS[pa_processTable[n1].initial].bddR;
    name1 = pa_processTable[n1].name;
  } else {
    D1 = pa_compositionTable[n1].transitionBDD;
    St1 = pa_compositionTable[n1].stateBDD;
    tau = pa_sortTable[pa_compositionTable[n1].sort].table[0].p;
    stab1 = pa_compositionTable[n1].stab;
    ini1 = pa_compositionTable[n1].initialBDD;
    name1 = pa_compositionTable[n1].name;
  }

  if (par2 == 0) {
    D2 = pa_processTable[n2].d;
    St2 = pa_processTable[n2].stateBDD;
    stab2 = pa_processTable[n2].stab;
    ini2 = pa_processTable[n2].tableS[pa_processTable[n2].initial].bddR;
    name2 = pa_processTable[n2].name;
  } else {
    D2 = pa_compositionTable[n2].transitionBDD;
    St2 = pa_compositionTable[n2].stateBDD;
    stab2 = pa_compositionTable[n2].stab;
    ini2 = pa_compositionTable[n2].initialBDD;
    name2 = pa_compositionTable[n2].name;
  }

  /* PRVI PROCES */

  /* spremenimo prvi proces */
  pom = Bdd_ExA(D1,tau);
  eps1 = VersisTrCl(pom);

  eps = Bdd_ITE(eps1,bdd_termTrue,stab1);
  Bdd_IncCounter();
  Bdd_Fresh(eps);

  tr1 = Bdd_RelOp(D1,bdd_termFalse,"R2P S2Q",TRUE);   /* preimenujemo sprem. stanja iz rs v pq */ 
  eps_rp = Bdd_S2P(eps);
  pom = Bdd_ExP(eps_rp,tr1);
  Bdd_IncCounter();
  Bdd_Fresh(eps);
  Bdd_Fresh(pom);

  eps_qs = Bdd_R2Q(eps);
  tr1_novi = Bdd_ExQ(pom,eps_qs);
  Bdd_IncCounter();
  Bdd_Fresh(tr1_novi);

  pom = Bdd_ITE(St1,stab1,bdd_termFalse);
  pom = Bdd_ITE(pom,tau,bdd_termFalse);
  tr1_novi = Bdd_ITE(tr1_novi,bdd_termTrue,pom);
  Bdd_Fortify(tr1_novi);
  Bdd_IncCounter();

  /* DRUGI PROCES */

  /* spremenimo drugi proces */
  pom = Bdd_ExA(D2,tau);
  eps2 = VersisTrCl(pom);

  eps = Bdd_ITE(eps2,bdd_termTrue,stab2);
  Bdd_IncCounter();
  Bdd_Fresh(eps);

  tr2 = Bdd_RelOp(D2,bdd_termFalse,"R2P S2Q",TRUE); /* preimenujemo sprem. stanja iz rs v pq */ 
  eps_rp = Bdd_S2P(eps);
  pom = Bdd_ExP(eps_rp,tr2);
  Bdd_IncCounter();
  Bdd_Fresh(eps);
  Bdd_Fresh(pom);

  eps_qs = Bdd_R2Q(eps);
  tr2_novi = Bdd_ExQ(pom,eps_qs);
  Bdd_IncCounter();
  Bdd_Fresh(tr2_novi);

  pom = Bdd_ITE(St2,stab2,bdd_termFalse);
  pom = Bdd_ITE(pom,tau,bdd_termFalse);
  tr2_novi = Bdd_ITE(tr2_novi,bdd_termTrue,pom);
  Bdd_Fortify(tr2_novi);
  Bdd_IncCounter();
 
  /* SHRANJEVANJE VMESNIH GRAFOV */

  /*
  Bdd_SaveFormula("Obs_",name1,tr1_novi);
  Bdd_SaveFormula("Obs_",name2,tr2_novi);
  */

  /* SIBKA OPAZOVALNA EKVIVALENCA */

  /* potrebno je preimenovanje drugega procesa */
  tr2_novi = Bdd_RelOp(tr2_novi,bdd_termFalse,"R2P S2Q",TRUE);
  St2 = Bdd_R2P(St2);
  Bdd_Fortify(tr2_novi);
  Bdd_Fortify(St2);
  Bdd_IncCounter();

  /* zacetni eqs */
  pom1 = Bdd_ExS(tr1_novi,bdd_termTrue);
  pom2 = Bdd_ExQ(tr2_novi,bdd_termTrue);
  pom3 = Bdd_ITE(pom1,Bdd_NOT(pom2),pom2);
  pom = Bdd_ExA(pom3,bdd_termTrue);

  eqs = Bdd_NOT(pom);
  eqs = Bdd_ITE(eqs,St1,bdd_termFalse);
  eqs = Bdd_ITE(eqs,St2,bdd_termFalse);
  Bdd_Fortify(eqs);
  Bdd_IncCounter();

  /* zacetno stanje */
  ini = Bdd_ITE(ini1,Bdd_R2P(ini2),bdd_termFalse);
  Bdd_Fortify(ini);

  EQV = Bisimulation(ini,eqs,St1,St2,tr1_novi,tr2_novi);

  return EQV;
}  

/**Function****************************************************************
  Synopsis    []
  Description [par1, par2 = 0 (process) or 1 (composition)]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Est_Boolean
VersisTestEquivalence(int par1, int n1, int par2, int n2)
{
  Bdd_Edge St1, St2, D1, D2;
  Bdd_Edge tau, stab1, stab2, init1, init2, Ptau1, Ptau2, trcl1, trcl2;
  Bdd_Edge Div1, Div2;
  Bdd_Edge pom, pom1, pom2, pom3, eps, eps_rp, eps_qs;
  int s;
  Bdd_Edge Acc1, Acc2, AccFalse1, AccFalse2, AccNull1, AccNull2;
  Bdd_Edge D1acc, D2acc;
  Bdd_Edge SR, QP, compat, compat0, compatNull;
  Bdd_Edge InitNew1, InitNew2, initNew, eqs;
  Est_Boolean EQV;
  Est_String inname1, inname2, name1, name2;

  /* dolocimo s, D, St, inname, tau, stab in init */
  /* sort in tau sta za oba procesa enaka! */

  if (par1 == 0) {
    D1 = pa_processTable[n1].d;
    inname1 = pa_processTable[n1].name;
    s = pa_processTable[n1].sort; /* sort */
    tau = pa_sortTable[s].table[0].p; /* tau */
    stab1 = pa_processTable[n1].stab;
    init1 = pa_processTable[n1].tableS[pa_processTable[n1].initial].bddR;
  } else {
    D1 = pa_compositionTable[n1].transitionBDD;
    inname1 = pa_compositionTable[n1].name;
    s = pa_compositionTable[n1].sort; /* sort */
    tau = pa_sortTable[s].table[0].p; /* tau */
    stab1 = pa_compositionTable[n1].stab;
    init1 = pa_compositionTable[n1].initialBDD;
  }

  if (par2 == 0) {
    D2 = pa_processTable[n2].d;
    inname2 = pa_processTable[n2].name;
    stab2 = pa_processTable[n2].stab;
    init2 = pa_processTable[n2].tableS[pa_processTable[n2].initial].bddR;
  } else {
    D2 = pa_compositionTable[n2].transitionBDD;
    inname2 = pa_compositionTable[n2].name;
    stab2 = pa_compositionTable[n2].stab;
    init2 = pa_compositionTable[n2].initialBDD;
  }

  /* PRVI PROCES */

  /* dolocimo Ptau1 - kam vse lahko gres iz posameznega stanja */
  /* eps so tau prehodi v D1 */

  pom = Bdd_ExA(D1,tau);
  trcl1 = VersisTrCl(pom);
  Bdd_Fortify(trcl1);
  Bdd_IncCounter();

  pom = Bdd_RelOp(D1,bdd_termFalse,"R2P S2Q",TRUE);
  eps = Bdd_ITE(trcl1,bdd_termTrue,stab1);
  Bdd_IncCounter();
  Bdd_Fresh(pom);
  Bdd_Fresh(eps);

  eps_rp = Bdd_S2P(eps);
  pom = Bdd_ExP(eps_rp,pom);          /* z leve mnozimo s tau prehodi */
  Bdd_IncCounter();
  Bdd_Fresh(eps);
  Bdd_Fresh(pom);

  eps_qs = Bdd_R2Q(eps);
  pom = Bdd_ExQ(pom,eps_qs);         /* z desne mnozimo s tau prehodi */
  Bdd_IncCounter();
  Bdd_Fresh(pom);

  Ptau1 = Bdd_ITE(tau,bdd_termFalse,pom);      /* brisemo tau prehode */
  Bdd_Fortify(Ptau1);
  Bdd_IncCounter();

  /* dolocimo divergentna stanja */
  Div1 = Divergent(D1,tau);
  Bdd_Fortify(Div1);
  Bdd_IncCounter();

  name1 = (Est_String) strdup("ACC_");
  name1 = (Est_String) realloc(name1,strlen(name1)+strlen(inname1)+1);
  strcat(name1,inname1);
  while (Pa_FindProcess(name1) >= 0) {
    name1 = (Est_String) realloc(name1,strlen(name1)+2);
    strcat(name1,"'");
  }

  D1acc = BuildAcc(par1,inname1,name1,D1,init1,Ptau1,trcl1,Div1,s,tau,
                   &Acc1,&AccFalse1,&AccNull1,&InitNew1);
  St1 = pa_processTable[Pa_FindProcess(name1)].stateBDD;

  /*
  printf("Acceptance graph %s is created.\n",name1);
  */

  /* DRUGI PROCES */

  /* dolocimo Ptau2 - kam vse lahko gres iz posameznega stanja */
  /* eps so tau prehodi v D2 */

  pom = Bdd_ExA(D2,tau);
  trcl2 = VersisTrCl(pom);
  Bdd_Fortify(trcl2);
  Bdd_IncCounter();

  pom = Bdd_RelOp(D2,bdd_termFalse,"R2P S2Q",TRUE);
  eps = Bdd_ITE(trcl2,bdd_termTrue,stab2);
  Bdd_IncCounter();
  Bdd_Fresh(pom);
  Bdd_Fresh(eps);

  eps_rp = Bdd_S2P(eps);
  pom = Bdd_ExP(eps_rp,pom);                /* mnozimo s tau prehodi */
  Bdd_IncCounter();
  Bdd_Fresh(eps);
  Bdd_Fresh(pom);

  eps_qs = Bdd_R2Q(eps);
  pom = Bdd_ExQ(pom,eps_qs);                /* mnozimo s tau prehodi */
  Bdd_IncCounter();
  Bdd_Fresh(pom);

  Ptau2 = Bdd_ITE(tau,bdd_termFalse,pom);     /* brisemo tau prehode */
  Bdd_Fortify(Ptau2);
  Bdd_IncCounter();

  /* dolocimo divergentna stanja */
  Div2 = Divergent(D2,tau);
  Bdd_Fortify(Div2);
  Bdd_IncCounter();

  name2 = (Est_String) strdup("ACC_");
  name2 = (Est_String) realloc(name2,strlen(name2)+strlen(inname2)+1);
  strcat(name2,inname2);
  while (Pa_FindProcess(name2) >= 0) {
    name2 = (Est_String) realloc(name2,strlen(name2)+2);
    strcat(name2,"'");
  }
  D2acc = BuildAcc(par2,inname2,name2,D2,init2,Ptau2,trcl2,Div2,s,tau,
                   &Acc2,&AccFalse2,&AccNull2,&InitNew2);
  St2 = pa_processTable[Pa_FindProcess(name2)].stateBDD;

  /*
  printf("Acceptance graph %s is created.\n",name2);
  */

  /* preimenovanje InitNew2, D2acc in St2 */
  InitNew2 = Bdd_R2P(InitNew2);
  D2acc = Bdd_RelOp(D2acc,bdd_termFalse,"R2P S2Q",TRUE);
  St2 = Bdd_R2P(St2);

  Bdd_Fortify(InitNew2);
  Bdd_Fortify(D2acc);
  Bdd_Fortify(St2);
  Bdd_IncCounter();

  /* compat0 */
  /* dobimo compat0 (r,p) */
  AccFalse2 =Bdd_R2P(AccFalse2);
  compat0 = Bdd_ITE(AccFalse1,AccFalse2,bdd_termFalse);

  /* compat */
  /* uporabljamo Acc1 (s,a,r) in Acc2 (q,a,p) */
  /* dobimo compat (r,p) */
  Acc2 = Bdd_RelOp(Acc2,bdd_termFalse,"R2P S2Q",TRUE);
  SR = Bdd_ExA(Acc1,bdd_termTrue);
  QP = Bdd_ExA(Acc2,bdd_termTrue);
  pom = Bdd_AxA(Acc1,Bdd_NOT(Acc2));    /* pom = AxA (Acc2 => Acc1) */
  pom = Bdd_ExP(pom,QP);                /* pom = ExP (QP * pom) */
  pom = Bdd_AxR(pom,Bdd_NOT(SR));       /* pom = AxR (SR => pom) */
  compat = pom;
  pom = Bdd_AxA(Acc2,Bdd_NOT(Acc1));    /* pom = AxA (Acc1 => Acc2) */
  pom = Bdd_ExR(pom,SR);                /* pom = ExR (SR * pom) */
  pom = Bdd_AxP(pom,Bdd_NOT(QP));       /* pom = AxP (QP => pom) */
  compat = Bdd_ITE(compat,pom,bdd_termFalse);
  compat = Bdd_RelOp(compat,bdd_termFalse,"S2R Q2P",TRUE);

  /* compatNull */
  /* dobimo compatNull (r,p) */
  AccNull2 =Bdd_R2P(AccNull2);
  compatNull = Bdd_ITE(AccNull1,AccNull2,bdd_termFalse);

  /* stanja z enakimi prehodi */
  /* uporabljamo D1acc (r,a,s) in D2acc (p,a,q) */
  /* dobimo eqs (r,p) */
  pom1 = Bdd_ExS(D1acc,bdd_termTrue);
  pom2 = Bdd_ExQ(D2acc,bdd_termTrue);
  pom3 = Bdd_ITE(pom1,Bdd_NOT(pom2),pom2);
  pom = Bdd_ExA(pom3,bdd_termTrue);
  eqs = Bdd_NOT(pom);

  /* celotna zacetna relacija */
  pom = Bdd_ITE(compat0,bdd_termTrue,compat);
  pom = Bdd_ITE(pom,bdd_termTrue,compatNull);
  eqs = Bdd_ITE(eqs,pom,bdd_termFalse);
  Bdd_Fortify(eqs);
  Bdd_IncCounter();

  /* zacetno stanje */
  initNew = Bdd_ITE(InitNew1,InitNew2,bdd_termFalse);
  Bdd_Fortify(initNew);
  Bdd_IncCounter();

  EQV = Bisimulation(initNew,eqs,St1,St2,D1acc,D2acc);

  return EQV;
}

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description [Transition relations must be D1(r,a,s) and D2(p,a,q)]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Est_Boolean
Bisimulation(Bdd_Edge init, Bdd_Edge eqs, Bdd_Edge St1, Bdd_Edge St2,
             Bdd_Edge D1, Bdd_Edge D2)
{
  Bdd_Edge eqs_old_rp, eqs1, eqs2, pom;

  /* ali sta zacetni stanji ekvivalentni */
  pom = Bdd_ITE(init,eqs,bdd_termFalse);
  if (Bdd_isEqv(pom,bdd_termFalse)) return FALSE;

  do {

    eqs_old_rp = eqs;                     /* eqs = eqs(r,p) */
    eqs = Bdd_RelOp(eqs,bdd_termFalse,"R2S P2Q",TRUE);

    /* Part 1 */
    pom = Bdd_ExQ(D2,eqs);                /* pom = pom(a,p,s) */
    eqs1 = Bdd_NewEq(D1,pom);             /* eqs1 = eqs1(r,p) */
    eqs1 = Bdd_ITE(eqs1,St1,bdd_termFalse);

    /* Part 2 */
    pom = Bdd_ExS(D1,eqs);                /* pom = pom(a,r,q) */
    eqs2 = Bdd_NewEq(D2,pom);             /* eqs2 = eqs2(r,p) */
    eqs2 = Bdd_ITE(eqs2,St2,bdd_termFalse);

    /* new eqs */
    eqs = Bdd_ITE(eqs1,eqs2,bdd_termFalse);
    eqs = Bdd_ITE(eqs,eqs_old_rp,bdd_termFalse);

    /* are initial states still equivalent? */
    pom = Bdd_ITE(init,eqs,bdd_termFalse);
    if (Bdd_isEqv(pom,bdd_termFalse)) {
      return FALSE;
    }

    Bdd_IncCounter();
    Bdd_Fresh(eqs);

  } while (!Bdd_isEqv(eqs_old_rp,eqs));

  /* TA IZPIS JE NAMENJEN MINIMIZACIJI */

  /*
  eqs = Bdd_RelOpComplex(eqs,bdd_termFalse,
                         "Ex xR<B1><B2> xP<B1><B2>",
                          TRUE);

  fprintf(stdout,"\nBisimulacija - eqs NA KONCU:\n");
  Pa_DecodeCompCompPair(&pa_compositionTable[pom_n1],
                        &pa_compositionTable[pom_n2],
                        eqs,TRUE);
  */

  return TRUE;
}

/**Function****************************************************************
  Synopsis    [Divergent]
  Description [Compute divergent states.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Edge
Divergent(Bdd_Edge D, Bdd_Edge tau)
{
  Bdd_Edge Div,last,new,pom;
  
  last = bdd_termFalse; 
  Div = bdd_termTrue; 

  while (!Bdd_isEqv(Div,last)) {
    Bdd_IncCounter();
    Bdd_Fresh(Div);

    last = Div;
    new = Bdd_ITE(D,Div,bdd_termFalse);
    pom = Bdd_R2S(Div);
    new = Bdd_ExS(new,pom);
    new = Bdd_ExA(new,tau);
    Div = new;
  }

  return Div;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Edge
BuildAcc(int par, Est_String name, Est_String newname,
         Bdd_Edge D, Bdd_Edge init, Bdd_Edge Ptau,
         Bdd_Edge trcl, Bdd_Edge Div, int s, Bdd_Edge tau,
         Bdd_Edge *Acc, Bdd_Edge *AccFalse, Bdd_Edge *AccNull,
         Bdd_Edge *InitNew)
{
  Pa_State *tabS;
  Bdd_Edge pom,pom1;
  int n,i;
  Est_Boolean b;

  /* tvorimo ogrodje novega procesa */
  Pa_AddNewProcess(newname);
  n = pa_processes-1;
  pa_processTable[n].sort = s;
  pa_processTable[n].numStates= 0;
  pa_processTable[n].initial = 0;

  ST = (StateTable *) malloc (sizeof(StateTable));
  STnum = 0;

  /* dolocimo zacetno stanje */
  pom = Bdd_ITE(init,trcl,bdd_termFalse);
  pom = Bdd_NewState(pom);
  init = Bdd_ITE(init,bdd_termTrue,pom);
  Bdd_Fortify(init);

  /* rekurzivna funkcija, ki zgradi acc proces */
  pom1 = Bdd_ITE(init,Div,bdd_termFalse);
  b = Bdd_isEqv(pom1,bdd_termFalse);
  BuildAccR(par,name,D,Ptau,init,tau,Div,b,s);

  /* encode process */
  Pa_EncodeProcess(newname);
  Bdd_IncCounter();

  /* initial state */
  *InitNew = pa_processTable[n].tableS[pa_processTable[n].initial].bddR;

  /* build acceptance relations */
  tabS = pa_processTable[n].tableS;
  *Acc = bdd_termFalse;
  *AccFalse = bdd_termFalse;
  *AccNull = bdd_termFalse;
  for (i=0; i<pa_processTable[n].numStates; i++) {
    if (Bdd_isNull(tabS[i].acc)) {
      *AccNull = Bdd_ITE(*AccNull,bdd_termTrue,tabS[i].bddR);
    } else {
      if (Bdd_isEqv(tabS[i].acc,bdd_termFalse)) {
        *AccFalse = Bdd_ITE(*AccFalse,bdd_termTrue,tabS[i].bddR);
      } else {
        pom = Bdd_ITE(tabS[i].bddS,tabS[i].acc,bdd_termFalse);
        *Acc = Bdd_ITE(*Acc,bdd_termTrue,pom);
      }
    }
  }

  Bdd_Fortify(*Acc);
  Bdd_Fortify(*AccFalse);
  Bdd_Fortify(*AccNull);
  Bdd_IncCounter();

  return pa_processTable[n].d;
}

/**Function****************************************************************
  Synopsis    []
   Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BuildAccR(int par, Est_String name,
          Bdd_Edge D, Bdd_Edge Ptau, Bdd_Edge S, Bdd_Edge tau, 
          Bdd_Edge Div, Est_Boolean b, int s)
{
  int n;
  Bdd_Edge pom, pom1, acc, a, Sa;
  Est_Boolean c;
  Est_String x;
  int j,t,tnew;

  /*
  if (b) printf("b TRUE\n"); else printf("b FALSE\n");
  */

  if (b) {
   acc = MakeAcc(D,tau,S);
   Bdd_Fortify(acc);
  } else {
   acc = bdd_termNull;
  }

  if (b) x = (Est_String) strdup("s"); else x = (Est_String) strdup("e");
  t = Pa_FOANextStateProcess(&pa_processTable[pa_processes-1],x);

  pa_processTable[pa_processes-1].tableS[t].acc = acc;
  AddStateTable(S,b,t);

  /* SLEDI PRIPRAVA ZA KONTROLNE IZPISE */

  if (par == 0) {
    n = Pa_FindProcess(name);
  }
  if (par == 1) {
    n = Pa_FindComposition(name);
  }

  /* za vsako akcijo, ki gre iz S dodamo prehode */
  /* akcija pa_sortTable[0] je akcija TAU, ki jo preskocimo! */

  for (j=1; j<pa_sortTable[s].numActions; j++) {
    a = pa_sortTable[s].table[j].p;

    /*
    printf("Action: %s\n",pa_sortTable[s].table[j].name);
    Bdd_PutsFunction(a);
    */

    /* output action */

    pom = Bdd_ITE(pa_sortTable[s].a0,a,bdd_termFalse);
    Sa = Bdd_ITE(S,pom,bdd_termFalse);
    Sa = Bdd_ITE(Sa,Ptau,bdd_termFalse);
    if (!Bdd_isEqv(Sa,bdd_termFalse)) {
      Sa = Bdd_NewState(Sa);
      Bdd_Fortify(Sa);
      pom1 = Bdd_ITE(Sa,Div,bdd_termFalse);
      c = (b && Bdd_isEqv(pom1,bdd_termFalse));

      /*
      printf("OUTPUT Sa:\n");
      Bdd_PutsFunction(Sa);
      */

      if (!FindStateTable(Sa,c,&tnew))
        tnew = BuildAccR(par,name,D,Ptau,Sa,tau,Div,c,s);
      Pa_FOATransition(&pa_processTable[pa_processes-1], t, j, 1, tnew);
    }

    /* input action */

    pom = Bdd_ITE(pa_sortTable[s].a0,bdd_termFalse,a);
    Sa = Bdd_ITE(S,pom,bdd_termFalse);
    Sa = Bdd_ITE(Sa,Ptau,bdd_termFalse);
    if (!Bdd_isEqv(Sa,bdd_termFalse)) {
      Sa = Bdd_NewState(Sa);
      Bdd_Fortify(Sa);
      pom1 = Bdd_ITE(Sa,Div,bdd_termFalse);
      c = (b && Bdd_isEqv(pom1,bdd_termFalse));

      /*
      printf("INPUT Sa:\n");
      Bdd_PutsFunction(Sa);
      */

      if (!FindStateTable(Sa,c,&tnew))
        tnew = BuildAccR(par,name,D,Ptau,Sa,tau,Div,c,s);
      Pa_FOATransition(&pa_processTable[pa_processes-1], t, j, 0, tnew);
    }
  }

  return t;
}

/**Function****************************************************************
  Synopsis    []
  Description [Determine acceptance set. The resulting set is not minimal!
               Special functions for comparing them should be used.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Edge
MakeAcc(Bdd_Edge D, Bdd_Edge tau, Bdd_Edge S)
{
  Bdd_Edge pom,pom1,eps,acc;

  pom = Bdd_ExS(S,D);
  pom1 = Bdd_ExA(pom,bdd_termTrue);

  if (Bdd_isEqv(pom1,S)) {
    eps = Bdd_ExA(pom,tau);
    acc = Bdd_ITE(eps,bdd_termFalse,pom);
  } else {
    acc = bdd_termFalse;
  }

  return acc;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
MinAcc(int numNS, Pa_State *tabNS, int numS, Pa_State *tabS)
{
  Bdd_Edge f,ri,pom;
  int i,j;

  for (i=0; i<numNS; i++) {
    f = tabNS[i].acc;
    if (!Bdd_isNull(f)) {
      if (!Bdd_isEqv(f,bdd_termFalse)) {
        for (j=0; j<numS; j++) {
          ri = tabS[j].bddR;
          pom = Bdd_ExR(f,ri);
          pom = Bdd_NOT(pom);
          if (!Bdd_isEqv(pom,bdd_termTrue)) {
            pom = Bdd_ExA(pom,f);
            pom = Bdd_ITE(pom,bdd_termTrue,ri);
          }
          f = Bdd_ITE(f,pom,bdd_termFalse);
        }
      }
      tabNS[i].minacc = f;
    }
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Est_Boolean
CmpAcc(Bdd_Edge acc1, Bdd_Edge acc2)
{
  Bdd_Edge pom,pom1;
  Bdd_Edge R,S;

  acc2 = Bdd_R2S(acc2);
  R = Bdd_ExA(acc1,bdd_termTrue);
  S = Bdd_ExA(acc2,bdd_termTrue);

  pom = Bdd_AxA(acc1,Bdd_NOT(acc2));    /* pom = AxA (acc2 => acc1) */
  pom = Bdd_ExS(pom,S);                 /* pom = ExS (S * pom) */
  pom = Bdd_AxR(pom,Bdd_NOT(R));       /* must = AxR (R => pom) */
  pom1 = pom;

  pom = Bdd_AxA(acc2,Bdd_NOT(acc1));    /* pom = AxA (acc1 => acc2) */
  pom = Bdd_ExR(pom,R);                 /* pom = ExS (R * pom) */
  pom = Bdd_AxS(pom,Bdd_NOT(S));       /* must = AxR (S => pom) */
  pom1 = Bdd_ITE(pom1,pom,bdd_termFalse);

  if (Bdd_isEqv(pom1,bdd_termTrue)) {
    return TRUE;
  } else {
    return FALSE;
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
AddStateTable(Bdd_Edge S, Est_Boolean b, int t)
{

  if (STnum == 0) {
    ST[0].S = S;
    ST[0].b = b;
    ST[0].t = t;
    STnum++;
  } else {
    STnum++;
    ST = (StateTable *) realloc (ST, STnum * sizeof(StateTable));
    ST[STnum-1].S = S;
    ST[STnum-1].b = b;
    ST[STnum-1].t = t;
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Est_Boolean
FindStateTable(Bdd_Edge S, Est_Boolean b, int *t)
{
  int i=0;

  while (i<STnum) {
    if ((Bdd_isEqv(ST[i].S,S)) && (ST[i].b == b)) {
      *t = ST[i].t;
      return TRUE;
    } else {
      i++;
    }
  }

  return FALSE;
}
